#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ..utilities import ensure_aid

from .. import TFmath
from .. import representations

from .algorithms import algorithms
from .algorithms import order_reduce_programs


def Q_rank_calc(z, p):
    if p.real == 0 or z.real == 0:
        if p.real == z.real:
            Q_rank = 0
        else:
            # TODO
            # should use the data spacing to regularize this case
            Q_rank = 1e3
    else:
        res_ratio = z.real / p.real
        Q_rank = abs(p - z) * (1 / (p.real) ** 2 + 1 / (z.real) ** 2) ** 0.5 + abs(
            res_ratio - 1 / res_ratio
        )
    return Q_rank


def Q_rank_single(r):
    Q_rank = abs(r.imag / r.real)
    return Q_rank


def order_reduce(
    aid,
    Q_rank_cutoff=1,
    optimize=True,
    reduce_c=True,
    reduce_r=False,
    hint_name=None,
):
    # TODO, can run a high Q_rank_cutoff and the progressively add pairs back in!
    #
    aid = ensure_aid(aid)

    Pc = aid.fitter.poles.c
    Zc = aid.fitter.zeros.c
    Pr = aid.fitter.poles.r
    Zr = aid.fitter.zeros.r

    removed_rzp_list = []
    # print(len(Zc), len(Pc))
    if reduce_c:
        rpB = TFmath.nearest_pairs(Zc, Pc)
        Zc = list(rpB.l1_remain)
        Pc = list(rpB.l2_remain)

        for z, p in rpB.r12_list:
            Q_rank = Q_rank_calc(p, z)
            # print(z, p, Q_rank)
            if Q_rank < Q_rank_cutoff:
                removed_rzp_list.append((Q_rank, z, p))
                continue
            Zc.append(z)
            Pc.append(p)

    if reduce_r:
        rpB = TFmath.nearest_pairs(Zr, Pr)
        Zr = list(rpB.l1_remain)
        Pr = list(rpB.l2_remain)

        for z, p in rpB.r12_list:
            Q_rank = Q_rank_calc(p, z)
            # print(z, p, Q_rank)
            if Q_rank < Q_rank_cutoff:
                removed_rzp_list.append((Q_rank, z, p))
                continue
            Zr.append(z)
            Pr.append(p)

    fitter_new = aid.fitter.regenerate(
        ZPKrep=aid.fitter.ZPKrep,
        z_c=Zc,
        p_c=Pc,
        z_r=Zr,
        p_r=Pr,
    )
    import numpy as np
    # These asserts should possibly be run only if ensure_stable is in play
    # shouldn't be so necessary after previous debugging
    # assert(np.all(fitter_new.poles.fullplane.real < 0))
    assert(fitter_new.order_relative == aid.fitter.order_relative)
    if optimize:
        with fitter_new.with_codings_only([fitter_new.gain_coding]):
            fitter_new.optimize()
        fitter_new.optimize()
        algorithms.sign_check_flip(fitter_new)
        aid.fitter_update(fitter_new)
    else:
        algorithms.sign_check_flip(fitter_new)
        aid.fitter_update(fitter_new, representative=False)
    
    # These asserts should possibly be run only if ensure_stable is in play
    # shouldn't be so necessary after previous debugging
    # assert(np.all(aid.fitter.poles.fullplane.real < 0))
    aid.fitter_checkpoint(hint_name=hint_name)
    return removed_rzp_list


def order_restore(
    aid,
    removed_rzp_list,
    hint_name=None,
):
    rank_zp_idx_list = []

    removed_rzp_list.sort()
    for r, z, p in removed_rzp_list:
        algorithms.resrank_program(
            aid.fitter,
            rank_zp_idx_list,
            "zp_restore",
            ["ZcAdd", z, "PcAdd", p],
            marginalize_delay=False,
        )

    did_reduce = True
    ever_restored = False
    while did_reduce:
        trials, rank_zp_idx_list = algorithms.ranking_reduction_trials(
            aid,
            rank_zp_idx_list,
            greedy=True,
            return_remaining=True,
            num_total_max=10,
            hint_name=hint_name,
        )
        if not trials:
            return False
        algorithms.sign_check_flip(trials[0].fitter)
        # print("HMM", aid.fitter.residuals_average, trials[0].fitter.residuals_average)
        trial = trials[0]
        did_reduce = aid.fitter_check(
            trial.fitter,
            hint_name="ordrestore",
            variant=trial.ord_str,
        )
        if did_reduce:
            ever_restored = True
    return ever_restored


def order_reduce_successive(
    aid,
    num_total_max=4,
    num_type_max=2,
    marginalize_delay=True,
    representative=True,
    hint_name=None,
):
    deg_min = aid.hint("total_degree_min")
    if deg_min is None:
        return
    last_flip_order = aid.fitter_orders().maxzp + 1
    while aid.fitter_orders().maxzp > deg_min:
        greedy_order = aid.hint("greedy_order")
        if greedy_order is None:
            greedy = False
        elif greedy_order < aid.fitter_orders().maxzp:
            greedy = True
        else:
            greedy = False

        #TODO
        greedy = False
        rank_zp_idx_list = []

        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_c(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_r(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_c2r(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_cc(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_rr(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_crr(
                aid, marginalize_delay=marginalize_delay
            )
        )
        # this one can infinite loop!
        if aid.fitter_orders().maxzp < last_flip_order:
            rank_zp_idx_list.extend(
                order_reduce_programs.ranking_reduction_r_flip(
                    aid, marginalize_delay=marginalize_delay
                )
            )
        rank_zp_idx_list.sort(key=lambda pb: pb.rank)
        trials = algorithms.ranking_reduction_trials(
            aid,
            rank_zp_idx_list,
            num_total_max=num_total_max,
            num_type_max=num_type_max,
            ranking_factor_max=100,
            greedy=greedy,
            hint_name=hint_name,
        )
        if not trials:
            return False
        trial = trials[0]
        algorithms.sign_check_flip(trial.fitter)
        # print("HMM", aid.fitter.residuals_average, trial.fitter.residuals_average)
        aid.fitter_update(
            trial.fitter,
            representative=representative,
        )

        # needs to be after the update to use the new maxzp
        is_flip = trial.pbunch.get('flip', False)
        if is_flip:
            last_flip_order = aid.fitter_orders().maxzp

        aid.log_progress(
            5,
            ("order reduced to {}, residuals={:.2e}").format(
                aid.fitter_orders().maxzp, aid.fitter.residuals_average
            ),
        )
        if not representative:
            aid.fitter_checkpoint(hint_name=hint_name)
    return


def order_reduce_selective(
    aid,
    num_total_max=4,
    num_type_max=2,
    marginalize_delay=True,
    representative=True,
    hint_name=None,
):
    # TODO, make num_trials a hint value
    ever_reduced = False
    did_reduce = True
    while did_reduce:
        greedy_order = aid.hint("greedy_order")
        if greedy_order is None:
            greedy = False
        elif greedy_order < aid.fitter_orders().maxzp:
            greedy = True
        else:
            greedy = False

        #TODO
        greedy = False

        rank_zp_idx_list = []

        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_c(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_r(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_c2r(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_cc(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_rr(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.extend(
            order_reduce_programs.ranking_reduction_crr(
                aid, marginalize_delay=marginalize_delay
            )
        )
        rank_zp_idx_list.sort(key=lambda pb: pb.rank)
        trials = algorithms.ranking_reduction_trials(
            aid,
            rank_zp_idx_list,
            num_total_max=num_total_max,
            num_type_max=num_type_max,
            ranking_factor_max=100,
            greedy=greedy,
            hint_name=hint_name,
        )
        if not trials:
            return False
        trial = trials[0]
        algorithms.sign_check_flip(trial.fitter)
        # print("HMM", aid.fitter.residuals_average, trial.fitter.residuals_average)
        did_reduce = aid.fitter_check(
            trial.fitter,
            hint_name=hint_name,
            variant="OrdDn",
        )
        if did_reduce:
            aid.fitter_update(representative=representative)
            aid.fitter_checkpoint(hint_name=hint_name)
            aid.log_progress(
                5,
                ("order reduced to {}, residuals={:.2e}, reldeg={}").format(
                    aid.fitter_orders().maxzp, aid.fitter.residuals_average, aid.fitter_orders().reldeg,
                ),
            )
            ever_reduced = True
        else:
            aid.log_progress(6, "order not reduced")
    return ever_reduced



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
import scipy
import scipy.linalg
import scipy.special

from wield.bunch import Bunch

from .data_filtfit_split import DataFiltFitSplitBase
from ..svd import SVD_SV


def abs_sq(x):
    return x.real ** 2 + x.imag ** 2


class RationalDiscFilterRB(DataFiltFitSplitBase):
    phase_missing = False

    def __build__(
        self,
        _args=None,
        npoles=None,
        nzeros=None,
        ZPK=((), (), 1),
        mindelay=False,
        stabilize=False,
        parent=None,
        poles_overlay=None,
        zeros_overlay=None,
        **kwargs
    ):
        if _args:
            raise RuntimeError("Only keyword arguments allowed")
        self.npoles = npoles
        self.nzeros = nzeros
        self.mindelay = mindelay
        self.stabilize = stabilize

        if parent is not None:
            if poles_overlay is None:
                poles_overlay = parent.poles_overlay
            if zeros_overlay is None:
                zeros_overlay = parent.zeros_overlay
        else:
            if poles_overlay is None:
                poles_overlay = ()
            if zeros_overlay is None:
                zeros_overlay = ()

        self.poles_overlay = poles_overlay
        self.zeros_overlay = zeros_overlay

        super(RationalDiscFilterRB, self).__build__(parent=parent, **kwargs)

        # first to prevent lookups that shouldn't happen
        self.zeros = ZPK[0]
        self.h_b_ref = 1
        self.dependencies_for("zeros", "b_vec")
        self.poles = ZPK[1]
        self.h_a_ref = 1
        self.dependencies_for("poles", "a_vec")
        self.gain = ZPK[2]

        @self.deco_generator(clear=False)
        def a_vec(self):
            """
            Setup so that the gain at the reference freq is ALWAYS 1
            """
            return self.poly.fromroots(self.poles).real

        @self.deco_setter(clear=False)
        def a_vec(self, val):
            gain = self.gain
            self.gain = gain / val[-1]
            self.h_a_ref = self.poly.val(self.Xp_ref_P, val) / val[-1]
            val = val / val[-1]
            if self.get_raw("stabilize"):
                poles = self.poly.roots(val)
                # use set raw so it won't update the gain redundantly/erroneously
                if np.any(abs(poles) > 1):
                    # only prevent using this vector if needed, stability is helped if we don't go convert between roots and vectors too much
                    # actually set it so that we get the stabilize action
                    self.poles = poles
                    return self.TAG_NO_SET
                else:
                    self.set_raw("poles", poles, setter=lambda s, x: x)

            self.dependencies("poles")
            return val

        @self.deco_generator(clear=False)
        def b_vec(self):
            """
            Setup so the gain is always specified at the ref
            """
            # TODO fix gain deps
            return self.get_raw("gain") * self.poly.fromroots(self.zeros).real

        @self.deco_setter(clear=False)
        def b_vec(self, val):
            val = np.asarray(val)
            gain = val[-1]
            self.gain = gain
            self.h_b_ref = self.poly.val(self.Xp_ref_Z, val) / val[-1]
            if self.get_raw("mindelay"):
                zeros = self.poly.roots(val)
                if np.any(abs(zeros) > 1):
                    # only prevent using this vector if needed, stability is helped if we don't go convert between roots and vectors too much
                    # actually set it so that we get the stabilize action
                    self.zeros = zeros
                    return self.TAG_NO_SET
                else:
                    # use set raw so it won't update the gain redundantly/erroneously
                    self.set_raw("zeros", zeros, setter=lambda s, x: x)
            # since they had to be computed anyway for the gain adjustment
            self.dependencies("zeros")
            return val

        @self.deco_generator(clear=False)
        def zeros(self):
            return self.poly.roots(self.b_vec)

        @self.deco_setter(clear=False)
        def zeros(self, val):
            # get previous
            val = np.asarray(val)
            if self.get_raw("mindelay"):
                val = self.root_stabilize(
                    val,
                    method="flip",
                    real_neg="ignore",
                    real_pos="ignore",
                )
            self.dependencies("b_vec")
            h_ref_now = self.poly.valfromroots(self.Xp_ref_Z, val)
            h_ref_prev = self.h_b_ref
            self.gain = self.get_raw("gain") * abs(h_ref_prev / h_ref_now)
            self.h_b_ref = h_ref_now
            return val

        @self.deco_generator(clear=False)
        def poles(self):
            return self.poly.roots(self.a_vec)

        @self.deco_setter(clear=False)
        def poles(self, val):
            val = np.asarray(val)
            if self.get_raw("stabilize"):
                val = self.root_stabilize(
                    val,
                    method="flip",
                    real_neg="ignore",
                    real_pos="ignore",
                )
            self.dependencies("a_vec")
            h_ref_now = self.poly.valfromroots(self.Xp_ref_P, val)
            h_ref_prev = self.h_a_ref
            self.gain = self.get_raw("gain") / abs(h_ref_prev / h_ref_now)
            self.h_a_ref = h_ref_now
            return val

        @self.deco_generator(clear=False)
        def zeros_full(self):
            return np.concatenate([self.zeros, self.zeros_overlay])

        @self.deco_generator(clear=False)
        def poles_full(self):
            return np.concatenate([self.poles, self.poles_overlay])

        @self.deco_setter(clear=False)
        def gain(self, val):
            # print("GAIN: ", val)
            return val

        @self.deco_generator(clear=False)
        def gain(self):
            raise RuntimeError("shouldn't compute")
            return self.b_vec[-1] / self.a_vec[-1]

        @self.deco_generator(clear=False)
        def V_b(self):
            return self.poly.vander(
                self.Xp_grid_Z, self.nzeros
            ) * self.zeros_phasing.reshape(-1, 1)

        @self.deco_generator(clear=False)
        def V_a(self):
            return self.poly.vander(
                self.Xp_grid_P, self.npoles
            ) * self.poles_phasing.reshape(-1, 1)

        @self.deco_generator(clear=False)
        def h_a(self):
            # TODO account for poles_phase_adjust
            self.dependencies("poles", "a_vec")
            X = self.Xp_grid_P
            pval = self.get_default("poles", None)
            # pval = self.poles
            if pval is not None:
                val = self.poly.valfromroots(X, pval)
                return val * self.Xn_grid_P ** len(pval)

            a_vec = self.get_default("a_vec", None)
            if a_vec is not None:
                X = self.Xp_grid_P
                val = self.poly.val(X, a_vec)
                return val * self.Xn_grid_P ** (len(a_vec) - 1)

            # otherwise compute it from the preferred poles representation
            pval = self.poles
            val = self.poly.valfromroots(X, pval)
            return val * self.Xn_grid_P ** len(pval)

        @self.deco_generator(clear=False)
        def h_b(self):
            # TODO account for zeros_phase_adjust
            self.dependencies("zeros", "b_vec", "gain")
            X = self.Xp_grid_Z
            zval = self.get_default("zeros", None)
            # zval = self.zeros
            if zval is not None:
                val = self.poly.valfromroots(X, zval)
                return self.gain * val * self.Xn_grid_Z ** len(zval)

            b_vec = self.get_default("b_vec", None)
            if b_vec is not None:
                # don't compute and normalize by h_ref since the b-coeffs carry the gain
                val = self.poly.val(X, b_vec)
                return val * self.Xn_grid_Z ** (len(b_vec) - 1)

            # otherwise compute it from the preferred poles representation
            zval = self.zeros
            val = self.poly.valfromroots(X, zval)
            return self.gain * val * self.Xn_grid_Z ** len(zval)

        @self.deco_generator(clear=False)
        def _extra_xfer_fit(self):
            if not self.zeros_overlay and not self.poles_overlay:
                return 1
            return self.poly.valfromroots(
                self.Xp_grid_Z, self.zeros_overlay
            ) / self.poly.valfromroots(self.Xp_grid_P, self.poles_overlay)

        @self.deco_generator(clear=False)
        def xfer_fit(self):
            return self._extra_xfer_fit * self.h_b / self.h_a

        @self.deco_generator(clear=False)
        def residuals(self):
            debias_reweight = 1 / (0.001 + self.W ** 2)
            retB = Bunch()
            R = self.xfer_fit / self.data
            retB.resP = self.W * (R - 1)
            retB.resZ = self.W * (1 / R - 1)
            retB.resD = self.W * (R - 1 / R)
            retB.resD_average = np.sum(abs_sq(retB.resD)) / (2 * len(self.data))
            retB.average = np.sum(
                (abs(retB.resP) ** 2 + abs(retB.resZ * debias_reweight) ** 2)
                / (1 + debias_reweight)
            ) / (2 * len(self.data))
            return retB

        @self.deco_generator(clear=False)
        def residuals_average(self):
            return self.residuals.resD_average

        self.zeros_phase_adjust = None
        self.poles_phase_adjust = None

        @self.deco_generator(clear=False)
        def zeros_phasing(self):
            if self.zeros_phase_adjust is None:
                return self.Xn_grid_Z ** self.nzeros
            elif self.zeros_phase_adjust < 0:
                return self.Xn_grid_Z ** (self.nzeros + self.zeros_phase_adjust)
            else:
                return self.Xn_grid_Z ** (self.zeros_phase_adjust)
            return

        @self.deco_generator(clear=False)
        def poles_phasing(self):
            if self.poles_phase_adjust is None:
                return self.Xn_grid_P ** self.npoles
            elif self.poles_phase_adjust < 0:
                return self.Xn_grid_P ** (self.npoles + self.poles_phase_adjust)
            else:
                return self.Xn_grid_P ** (self.poles_phase_adjust)
            return

        @self.deco_generator(clear=False)
        def A_z(self):
            return self.V_b * (self.W / (self.data * self.h_a)).reshape(-1, 1)

        @self.deco_generator(clear=False)
        def A_zp(self):
            return self.V_a * (self.W * self.h_b / (self.data * self.h_a ** 2)).reshape(
                -1, 1
            )

        @self.deco_generator(clear=False)
        def A_p(self):
            return self.V_a * (self.W * (self.data / self.h_b)).reshape(-1, 1)

        @self.deco_generator(clear=False)
        def A_pz(self):
            return self.V_b * (self.W * (self.data * self.h_a / self.h_b ** 2)).reshape(
                -1, 1
            )

        return  # ~__init__

    @property
    def ZPK(self):
        # so cheap to make that we don't try to autodelete
        # TODO, fix gain term as it is currently based on reference frequency
        return (self.zeros, self.poles, self.gain)

    def root_stabilize(
        self,
        roots,
        method=None,
        real_neg=None,
        real_pos=None,
    ):
        roots = np.asarray(roots)
        select = abs(roots) > 1
        if real_neg is None:
            pass
        elif real_neg == "ignore":
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real < 0))
        elif real_neg == "ignore_float":
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real < 0))
            float_sel = (
                (abs(roots.imag) < 1e-12) & (roots.real < -1) & (roots.real > -5)
            )
            roots[float_sel] = -5
        else:
            raise RuntimeError("Bad Argument")

        if real_pos is None:
            pass
        elif real_pos == "ignore":
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real > 0))
        elif real_pos == "ignore_float":
            select = select & ~((abs(roots.imag) < 1e-12) & (roots.real > 0))
            float_sel = (abs(roots.imag) < 1e-12) & (roots.real > 1) & (roots.real < 5)
            roots[float_sel] = 5
        else:
            raise RuntimeError("Bad Argument")

        if method is None:
            return roots
        elif method == "flip":
            roots = np.copy(roots)
            roots[select] = 1 / roots[select].conjugate()
            return roots
        elif method == "remove":
            # just nullify them
            roots = roots[~select]
            return roots
        elif method == "zero":
            roots[select] = 0
        else:
            raise RuntimeError("Bad Argument")
        return roots

    def clear_unstable_poles(self):
        new_poles = []
        for r in self.poles:
            if abs(r) < 1:
                new_poles.append(r)
        self.poles = new_poles

    def clear_bigdelay_zeros(self):
        new_zeros = []
        for r in self.zeros:
            if abs(r) < 1:
                new_zeros.append(r)
        self.zeros = new_zeros

    def matched_pairs_clear(
        self,
        Q_rank_cutoff=0.5,
    ):
        """
        Match unique closest pairs, if they are within a bandwidth of 0Hz, then they are ignored
        """
        poles_r, poles_c = self.poles_split
        zeros_r, zeros_c = self.zeros_split

        def nearest_idx(lst_1, lst_2):
            nearest_lst = []
            for r1 in lst_1:
                if r1 is None:
                    nearest_lst.append(None)
                    continue
                dist_nearest = float("inf")
                idx_nearest = None
                for idx_2, r2 in enumerate(lst_2):
                    if r2 is None:
                        continue
                    dist = abs(r1 - r2)
                    if dist < dist_nearest:
                        idx_nearest = idx_2
                        dist_nearest = dist
                nearest_lst.append(idx_nearest)
            return nearest_lst

        z_nearest = nearest_idx(zeros_c, poles_c)
        p_nearest = nearest_idx(poles_c, zeros_c)

        z_duals = []
        p_duals = []
        for idx_z, idx_p in enumerate(z_nearest):
            if idx_p is None:
                continue
            if idx_z != p_nearest[idx_p]:
                # not a unique pairing
                continue
            z = zeros_c[idx_z]
            p = poles_c[idx_p]
            Q_rank = abs(p - z) * (1 / (1 - abs(p)) ** 2 + 1 / (1 - abs(z)) ** 2) ** 0.5
            if Q_rank < Q_rank_cutoff:
                z_duals.append(idx_z)
                p_duals.append(idx_p)
        p_duals = set(p_duals)
        z_duals = set(z_duals)

        poles_new = []
        for idx_p, pole in enumerate(poles_c):
            if idx_p in p_duals:
                continue
            poles_new.append(pole)
            poles_new.append(pole.conjugate())
        zeros_new = []
        for idx_z, zero in enumerate(zeros_c):
            if idx_z in z_duals:
                continue
            zeros_new.append(zero)
            zeros_new.append(zero.conjugate())
        poles_new.extend(poles_r)
        zeros_new.extend(zeros_r)
        self.poles = poles_new
        self.zeros = zeros_new
        return

    def fit_poles_mod_zeros(self, **kwargs):
        # remove the effect of linear (delay) phasing
        A_z = self.A_z
        # A_z = np.hstack([(1j*F_Hz).reshape(-1, 1), A_z])
        q, r = np.linalg.qr(A_z)

        A_p = self.A_p
        # the str is because py2 doesn't like some symbol in there
        A_p = A_p - np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_p)
        )

        S, V = SVD_SV(
            np.vstack([A_p.real, A_p.imag]),
            n_smallest=3,
            overwrite_a=True,
        )
        # print("POLES SVD: ", S[-4:])
        self.a_vec = V.T[:, -1]
        return

    def fit_SVD(self, modify=True, residuals=0, **kwargs):
        # remove the effect of linear (delay) phasing
        A_z = self.A_z
        A_p = self.A_p
        if residuals == 0:
            pass
        elif residuals > 0:
            A_z = self.V_b * (self.W * self.zeros_phasing / self.h_b).reshape(-1, 1)
        elif residuals < 0:
            A_p = self.V_a * (self.W * self.poles_phasing / self.h_a).reshape(-1, 1)

        A_svd = np.hstack(
            [np.vstack([A_p.real, A_p.imag]), -np.vstack([A_z.real, A_z.imag])]
        )

        S, V = SVD_SV(
            A_svd,
            n_smallest=3,
            overwrite_a=True,
        )
        S_rescaled = abs(S / V.T[0, :])
        idx_best = np.argmin(S_rescaled)
        if modify:
            self.a_vec = V.T[: A_p.shape[1], idx_best]
            self.b_vec = V.T[A_p.shape[1] :, idx_best]
        return S, V.T[0, :]

    def fit_zeros_mod_poles(self, **kwargs):
        A_p = self.A_p
        # remove the effect of linear (delay) phasing
        # A_p = np.hstack([(1j*F_Hz).reshape(-1, 1), A_p])

        q, r = np.linalg.qr(A_p)

        A_z = self.A_z
        A_z = A_z - np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_z)
        )
        S, V = SVD_SV(
            np.vstack([A_z.real, A_z.imag]),
            n_smallest=3,
            overwrite_a=True,
        )
        # print("ZEROS SVD: ", S[-4:])
        self.b_vec = V.T[:, -1]
        return

    def fit_poles(self, **kwargs):
        # print(self.a_vec, self.b_vec)
        A_p = self.A_p
        # solve the problem with purely real taps
        a_fit, res, rank, s = scipy.linalg.lstsq(
            np.vstack([A_p.real, A_p.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        # print(a_fit)
        print(a_fit.shape)
        self.a_vec = a_fit
        return

    def fit_zeros(self, **kwargs):
        A_z = self.A_z
        b_fit, res, rank, s = scipy.linalg.lstsq(
            np.vstack([A_z.real, A_z.imag]),
            np.hstack([self.W.real, self.W.imag]),
        )
        self.b_vec = b_fit
        return

    def match_pair_iter(
        self,
        Q_rank_cutoff,
        num_sequence=2,
        zeros_first=True,
    ):
        npoles = self.npoles
        nzeros = self.nzeros

        self.matched_pairs_clear(Q_rank_cutoff=Q_rank_cutoff)

        self.npoles = len(self.poles)
        self.nzeros = len(self.zeros)

        if not zeros_first:
            self.fit_poles()
            num_sequence -= 1
        for idx in range(num_sequence):
            self.fit_zeros()
            self.fit_poles()

        if not zeros_first:
            self.fit_zeros()

        self.npoles = npoles
        self.nzeros = nzeros

    def fit_poles_s(self, **kwargs):
        A_p = self.A_p
        pt = np.exp(-1j * np.pi * np.linspace(self.F_Hz[-1], self.Fp_nyquist_Hz, 500))
        # pt = -1
        V_p = self.poly.vander(pt, self.npoles)
        prev = np.dot(V_p, list(self.a_vec) + [0] * (1 + self.nzeros - len(self.a_vec)))
        V_p = V_p / prev.reshape(-1, 1)
        W = 10
        a_fit, res, rank, s = scipy.linalg.lstsq(
            np.block([[A_p.real], [A_p.imag], [W * V_p.real], [W * V_p.imag]]),
            np.hstack(
                [self.W.real, self.W.imag] + [W * np.ones(500)] + [np.zeros(500)]
            ),
        )
        self.a_vec = a_fit[:]
        return

    def fit_poles2(self, **kwargs):
        # print(self.a_vec, self.b_vec)
        A_p = self.A_p
        # solve the problem with purely real taps
        print(A_p.real.shape)
        a_fit, res, rank, s = scipy.linalg.lstsq(
            # np.block([[A_p.real, 0 * self.W.reshape(-1, 1)], [A_p.imag, (self.W * 1j * self.F_Hz).reshape(-1, 1)]]),
            np.block(
                [
                    [A_p.real, 0 * self.W.reshape(-1, 1)],
                    [A_p.imag, (self.W * self.F_Hz).reshape(-1, 1)],
                ]
            ),
            np.hstack([self.W.real, self.W.imag]),
        )
        print(a_fit.shape)
        self.a_vec = a_fit[:-1]
        return

    def fit_zeros2(self, **kwargs):
        A_z = self.A_z
        b_fit, res, rank, s = scipy.linalg.lstsq(
            np.block(
                [
                    [A_z.real, 0 * self.W.reshape(-1, 1)],
                    [A_z.imag, (self.W * self.F_Hz).reshape(-1, 1)],
                ]
            ),
            np.hstack([self.W.real, self.W.imag]),
        )
        self.b_vec = b_fit[:-1]
        return

    def fit_zeros_s(self, **kwargs):
        A_z = self.A_z
        pt = np.exp(-1j * np.pi * np.linspace(self.F_Hz[-1], self.Fz_nyquist_Hz, 500))
        # pt = -1
        V_z = self.poly.vander(pt, self.nzeros)
        prev = np.dot(V_z, list(self.b_vec) + [0] * (1 + self.nzeros - len(self.b_vec)))
        V_z = V_z / prev.reshape(-1, 1)
        W = 10
        print(V_z.shape)
        b_fit, res, rank, s = scipy.linalg.lstsq(
            np.block([[A_z.real], [A_z.imag], [W * V_z.real], [W * V_z.imag]]),
            np.hstack(
                [self.W.real, self.W.imag] + [W * np.ones(500)] + [np.zeros(500)]
            ),
        )
        self.b_vec = b_fit[:]
        return

    def fit_zeros_s2(self, **kwargs):
        A_z = self.A_z

        Xp_grid = np.exp(1j * np.pi * self.F_Hz / np.max(self.F_Hz))
        V_b2 = self.poly.vander(Xp_grid, self.nzeros)
        z_phasing = Xp_grid ** -self.nzeros

        A_z2 = V_b2 * (self.W * z_phasing / (self.data * self.h_a)).reshape(-1, 1)

        q, r = np.linalg.qr(A_z2)

        A_z_proj = np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_z)
        )
        W_proj = np.einsum(
            str("ij,j->i"), q, np.einsum(str("ij,i->j"), q.conjugate(), self.W)
        )

        b_fit, res, rank, s = scipy.linalg.lstsq(
            np.block([[A_z_proj.real], [A_z_proj.imag]]),
            np.hstack([W_proj.real, W_proj.imag]),
        )
        self.b_vec = b_fit
        return

    def fit_poles_s2(self, **kwargs):
        A_p = self.A_p

        Xp_grid = np.exp(1j * np.pi * self.F_Hz / np.max(self.F_Hz))
        V_a2 = self.poly.vander(Xp_grid, self.npoles)
        p_phasing = Xp_grid ** -self.npoles

        A_p2 = V_a2 * (self.W * p_phasing * (self.data / self.h_b)).reshape(-1, 1)

        q, r = np.linalg.qr(A_p2)

        A_p_proj = np.einsum(
            str("ij,jk->ik"), q, np.einsum(str("ij,ik->jk"), q.conjugate(), A_p)
        )
        W_proj = np.einsum(
            str("ij,j->i"), q, np.einsum(str("ij,i->j"), q.conjugate(), self.W)
        )

        a_fit, res, rank, s = scipy.linalg.lstsq(
            np.block([[A_p_proj.real], [A_p_proj.imag]]),
            np.hstack([W_proj.real, W_proj.imag]),
        )
        self.a_vec = a_fit
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
Some notes about writing/reading a common dictionary format across these files:
Some of the issues can only be resolved by code-side normalization functions.

Only dictionaries and arrays of numbers should be supported
 - hdf5 cannot have heterogeneous lists of objects (others can)
 - matlab array squeezing can truncate 1x1 arrays to single elements. Only
   code-side normalization can fix this
 - json/yaml must split complex numbers into dicts with real/imag parts. This
   could/should be reversed on load, but currently requires code normalization
   to reverse
"""

import collections
import numpy as np
import copy

from . import types


def load_any(
    fname,
    ftype,
):
    if ftype == "hdf5":
        from . import hdf5_io

        fdict = hdf5_io.load_hdf5(fname)
    elif ftype == "json":
        from . import json_io

        fdict = json_io.load_json(fname)
    elif ftype == "yaml":
        from . import yaml_io

        fdict = yaml_io.yaml_load(fname)
    elif ftype == "pickle":
        from . import pickle_io

        fdict = pickle_io.load_pickle(fname)
    elif ftype == "ini":
        from . import ini_io

        fdict = ini_io.ini_load(fname)
    elif ftype == "mat":
        from . import matlab_io

        fdict = matlab_io.load_matlab(fname)
    elif ftype == "special":
        try:
            import IIRrational_test_data

            IIRrational_data = IIRrational_test_data.IIRrational_data_dev
        except (ImportError, AttributeError):
            from ..testing import IIRrational_data
        fdict = IIRrational_data(fname)
    else:
        raise RuntimeError("Unsupported filetype")
    return fdict


def cull_None(obj):
    if isinstance(obj, collections.Mapping):
        dels = []
        for k, v in obj.items():
            v2 = cull_None(v)
            if v2 is None:
                dels.append(k)
            else:
                obj[k] = v2
        for k in dels:
            del obj[k]
        return obj
    elif isinstance(obj, np.ndarray):
        if obj.dtype == object:
            for idx, v in np.ndenumerate(obj):
                obj[idx] = cull_None(v)
            return obj
        else:
            return obj
    elif isinstance(obj, list):
        for idx, v in enumerate(obj):
            obj[idx] = cull_None(v)
        return obj
    elif isinstance(obj, tuple):
        obj2 = [None] * len(obj)
        for idx, v in enumerate(obj):
            obj2[idx] = cull_None(v)
        return tuple(obj2)
    return obj


def normalize_ndarray(obj):
    if isinstance(obj, collections.Mapping):
        for k, v in obj.items():
            obj[k] = normalize_ndarray(v)
        return obj
    elif isinstance(obj, np.ndarray):
        return obj
    elif isinstance(obj, np.generic):
        # captures scalars given the previous if
        return np.asscalar(obj)
    elif isinstance(obj, (list, tuple)):
        return np.asarray(obj)
    return obj


def fix_complex(obj):
    normalize_ndarray(obj)
    if isinstance(obj, collections.Mapping):
        for k, v in obj.items():
            obj[k] = fix_complex(v)
        return obj
    elif isinstance(obj, np.ndarray):
        if np.iscomplexobj(obj):
            objD = collections.OrderedDict()
            objD["real"] = obj.real
            objD["imag"] = obj.imag
            objD["__type__"] = "[complex]"
            obj = objD
            return obj
        elif obj.dtype == object:
            for idx, v in np.ndenumerate(obj):
                obj[idx] = fix_complex(v)
            return obj
        else:
            return obj
    elif isinstance(obj, list):
        for idx, v in enumerate(obj):
            obj[idx] = fix_complex(v)
        return obj
    elif isinstance(obj, tuple):
        obj2 = [None] * len(obj)
        for idx, v in enumerate(obj):
            obj2[idx] = fix_complex(v)
        return tuple(obj2)
    elif isinstance(obj, complex):
        objD = collections.OrderedDict()
        objD["real"] = obj.real
        objD["imag"] = obj.imag
        objD["__type__"] = "[complex]"
        obj = objD
    return obj


def fix_ndarray(obj):
    if isinstance(obj, collections.Mapping):
        for k, v in obj.items():
            obj[k] = fix_ndarray(v)
        return obj
    elif isinstance(obj, np.ndarray):
        if obj.dtype == object:
            obj = fix_ndarray(obj.tolist())
            return obj
        else:
            obj = obj.tolist()
            return obj
    elif isinstance(obj, np.generic):
        return np.asscalar(obj)
    elif isinstance(obj, list):
        for idx, v in enumerate(obj):
            obj[idx] = fix_ndarray(v)
        return obj
    elif isinstance(obj, tuple):
        obj2 = [None] * len(obj)
        for idx, v in enumerate(obj):
            obj2[idx] = fix_ndarray(v)
        return tuple(obj2)
    return obj


def write_any(
    fname,
    ftype,
    fdict,
):
    features = types.type2features[ftype]
    fdict_orig = None

    if fdict_orig is None:
        fdict_orig = fdict
        fdict = copy.deepcopy(fdict)
    cull_None(fdict)
    if not features["complex"]:
        if fdict_orig is None:
            fdict_orig = fdict
            fdict = copy.deepcopy(fdict)
        fix_complex(fdict)

    if not features["ndarray"]:
        if fdict_orig is None:
            fdict_orig = fdict
            fdict = copy.deepcopy(fdict)
        fix_ndarray(fdict)

    if ftype == "hdf5":
        from . import hdf5_io

        fdict = hdf5_io.write_hdf5(fname, fdict)
    elif ftype == "mat":
        from . import matlab_io

        fdict = matlab_io.write_matlab(fname, fdict)
    elif ftype == "json":
        from . import json_io

        fdict = json_io.write_json(fname, fdict)
    elif ftype == "yaml":
        from . import yaml_io

        fdict = yaml_io.yaml_write(fname, fdict)
    elif ftype == "pickle":
        from . import pickle_io

        fdict = pickle_io.write_pickle(fname, fdict)
    elif ftype == "ini":
        from . import ini_io

        fdict = ini_io.ini_write(fname, fdict)
    elif ftype == "csv":
        raise RuntimeError("Unsupported Output Type")
    return fdict



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals

import numpy as np

from wield.iirrational.fitters_ZPK.MRF import check_jac
from wield.iirrational import testing


def check_residuals(fitterMRF):
    # axB = mplfigB()
    for idx in reversed(range(len(fitterMRF.parameters))):
        coding = fitterMRF.coding_maps.p2c[idx]
        # this indicates that the derivative is virtualized
        if coding.derivative_deadzoned:
            continue
        # print(coding, idx)
        try:
            val, select = check_jac(fitterMRF, idx, scaling=0.01)
            if np.any(~np.isfinite(val[select])):
                print(idx, fitterMRF.residuals_jacobian[idx])
                print(idx, val)
            testing.assert_almost_equal(val[select], 1, 2)
        except AssertionError:
            try:
                val, select = check_jac(fitterMRF, idx, scaling=0.0001)
                testing.assert_almost_equal(val[select], 1, 2)
            except AssertionError:
                coding = fitterMRF.coding_maps.p2c[idx]
                print(
                    idx,
                    coding,
                    coding.roots(),
                    coding.reduce(),
                )
                raise
        # axB.ax0.plot(fitterMRF.F_Hz, abs(val))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
import numpy as np
from wield.bunch import Bunch

# TODO, package with lib


# from .arguments import ArgumentError
# from .. import fitters_ZPK

from ..external.tabulate import tabulate
from .. import plots
from . import results_aid


class ResultsAidAdv(results_aid.ResultsAid):
    def __init__(self, fit_aid, **kwargs):
        super(ResultsAidAdv, self).__init__(fit_aid, **kwargs)
        self._investigations = {}
        self._investigations.update(fit_aid.investigations)
        for name, func in self._investigations.items():
            setattr(
                self,
                name,
                (lambda func: lambda *args, **kwargs: func(self, *args, **kwargs))(
                    func
                ),
            )
        return

    def investigate_fit_plot(self, fname=None, xscale="log_zoom", **kwargs):
        """
        Generates a plot of the fit. Returns a Bunch object which
        stores the axes and figure of the plot.
        """
        # axB = plots.plot_fitter_flag(fitter=self.fitter, xscale=xscale, **kwargs)
        axB = plots.plot_fitter_flag_residuals(fitter=self.fitter, xscale=xscale, **kwargs)
        if fname is not None:
            axB.save(fname)
        return axB

    def investigate_order_arrays(self, local_best=True):
        obo, rbo, fbo, lbo = self._fitters_by_order

        # only returns or displays those which are the best for that order or below
        if local_best:
            obo = obo[lbo]
            rbo = rbo[lbo]
            fbo = [fbo[idx] for idx in lbo]

        rmaxbo = []
        rmedbo = []
        for fit in fbo:
            rmaxbo.append(fit.residuals_max)
            rmedbo.append(fit.residuals_med)
        return Bunch(
            order=obo,
            res_avg=rbo,
            res_max=rmaxbo,
            res_med=rmedbo,
        )

    def investigate_order_console(
        self,
        tablefmt="simple",
        print_function=print,
    ):
        oB = self.investigate_order_arrays()
        pval = tabulate(
            np.asarray([oB.order, oB.res_avg, oB.res_med, oB.res_max]).T,
            ["max(z, p)\norder", "ChiSq.\navg. res.", "\nmed. res.", "\nmax. res."],
            tablefmt=tablefmt,
        )
        if print_function is not None:
            print_function(pval)
        else:
            return pval

    def investigate_order_plot(self, fname=None, xscale="log_zoom", **kwargs):
        obo, rbo, fbo, lbo = self._fitters_by_order
        axB = plots.plot_fitter_flag_residuals(
            fitter=self.fitter, xscale=xscale, **kwargs
        )
        for idx in reversed(range(len(fbo))):
            fit = fbo[idx]
            if fit is self.fitter:
                continue
            if not lbo[idx]:
                continue
            axB.plot_fit(
                fit,
                label=("order: {}, ChiSq: {:.3e}").format(
                    fit.order, fit.residuals_average
                ),
            )
        if fname is not None:
            axB.save(fname)
        else:
            axB.finalize()
        return axB

    def investigate_covarianceRI(self, output="print"):
        """
        outputs the covariance matrix for the real/imaginary parts of the chosen
        filter.
        """
        # TODO
        raise NotImplementedError()

    def investigate_valuesRI_print(self, output="print"):
        """
        outputs the covariance matrix for the real/imaginary parts of the chosen
        filter.

        output specifies the output format, which may be one of:
        - "print":
        - "arguments":
        - "matrix":
        - "values":
        """
        # TODO
        raise NotImplementedError()

    def investigate_fisherRI_print(self, output="print"):
        """
        outputs the covariance matrix for the real/imaginary parts of the chosen
        filter.

        output specifies the output format, which may be one of:
        - "print":
        - "arguments":
        - "matrix":
        - "values":
        """
        # TODO
        raise NotImplementedError()

    # def digest(
    #    self,
    #    fname = None,
    #    folder = None,
    #    verbosity = None,
    #    **kwargs
    # ):
    #    """
    #    Generates a document with the logging output, intermediate fits and
    #    investigations related to the fitting call. This is verbose and slow,
    #    but useful.

    #    The default is to print and plot to the output,
    #    if running from an ipython/jupyter notebook.
    #    """
    #    raise NotImplementedError()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import os
import re
import errno
import os.path as path

from ..utilities.mpl import asavefig
from ..utilities.strings import padding_remove


def header_line(level=0, section_history=(), heading=""):
    if not heading:
        heading = ""
    section_link = ".".join([str(n) for n in section_history])
    if section_link:
        section_link = " `{0}` ".format(section_link)
    else:
        section_link = " "
    return ("#" * (level + 1)) + section_link + heading


def docdb_md_section(
    F_md,
    doc_db,
    section_history,
    heading_level,
    md_verbosity=None,
    plot_verbosity=None,
    plot_fname_func=None,
    do_plot=True,
    regex_secname="",
    use_ext=True,
    plot_format="png",
):
    if md_verbosity is not None:
        if doc_db.verbosity:
            if doc_db.verbosity > md_verbosity:
                return

    header = header_line(
        heading_level,
        section_history,
        doc_db.name,
    )

    lines = [
        header,
        "",
    ]

    if doc_db.method:
        lines.append("> `{0}`".format(padding_remove(doc_db.method)))
        lines.append("")

    if doc_db.about:
        lines.append(padding_remove(doc_db.about))
        lines.append("")

    if not do_plot:
        if re.findall(regex_secname, doc_db.name):
            do_plot = True

    will_plot = True
    if plot_verbosity is not None:
        if doc_db.verbosity:
            if doc_db.verbosity > plot_verbosity:
                will_plot = False

    if do_plot and will_plot and doc_db.plotter and doc_db.fitter:
        title = doc_db.plot_title
        if not title:
            title = ""
        plot_fname = plot_fname_func(
            section_history,
            doc_db,
            with_folder=False,
        )
        if use_ext:
            ext = "." + plot_format
        else:
            ext = ""
        lines.append("![{title}]({0})".format(plot_fname + ext, title=title))
        lines.append("")

    F_md.writelines((l + "\n" for l in lines))

    if doc_db.sequence:
        for idx, seq_name in enumerate(doc_db.sequence):
            docdb_md_section(
                F_md,
                doc_db[seq_name],
                section_history=section_history + [idx + 1],
                heading_level=heading_level + 1,
                md_verbosity=md_verbosity,
                plot_verbosity=plot_verbosity,
                plot_fname_func=plot_fname_func,
                do_plot=do_plot,
                regex_secname=regex_secname,
                plot_format=plot_format,
                use_ext=use_ext,
            )
    return


def docdb_md_plots(
    doc_db,
    section_history,
    plot_verbosity=None,
    plot_fname_func=None,
    do_plot=True,
    regex_secname="",
    clear_plots=False,
    plot_format="png",
    dpi=100,
):
    if plot_verbosity is not None:
        if doc_db.verbosity:
            if doc_db.verbosity > plot_verbosity:
                return

    if not do_plot:
        if re.findall(regex_secname, doc_db.name):
            do_plot = True

    if do_plot and doc_db.plotter and doc_db.fitter:
        plot_fname = plot_fname_func(
            section_history,
            doc_db,
            with_folder=True,
        )
        print("PLOTTING: ", plot_fname)
        Fb = doc_db.plotter(doc_db.fitter)
        Fb.formats[plot_format].use = True
        Fb.formats[plot_format].facecolorize = True
        Fb.formats[plot_format].dpi = dpi
        Fb.save_show = False
        Fb.save(plot_fname, fixname=False)
    elif clear_plots:
        plot_fname = plot_fname_func(
            section_history,
            doc_db,
            with_folder=True,
        )
        print("REMOVING: ", plot_fname)
        try:
            os.remove(plot_fname + "." + plot_format)
        except OSError as e:
            if e.errno != errno.ENOENT:
                print(e)

    if doc_db.sequence:
        for idx, seq_name in enumerate(doc_db.sequence):
            docdb_md_plots(
                doc_db[seq_name],
                section_history=section_history + [idx + 1],
                plot_verbosity=plot_verbosity,
                plot_fname_func=plot_fname_func,
                do_plot=do_plot,
                regex_secname=regex_secname,
                clear_plots=clear_plots,
                plot_format=plot_format,
                dpi=dpi,
            )
    return


def docdb2markdown(
    folder_name,
    doc_db_root,
    md_name="digest.md",
    include_plots=True,
    include_md=True,
    plot_verbosity=None,
    md_verbosity=None,
    clear_plots=False,
    regex_plotsections=None,
    ipy_display=False,
    use_ext=True,
    plot_format="png",
    dpi=100,
    exporter=None,
    MP_workers=1,
):
    try:
        os.makedirs(folder_name)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    def plot_fname_func(
        section_history,
        doc_db,
        with_folder=True,
    ):
        section_link = "-".join([str(n) for n in section_history])
        if section_link:
            fname = "plot-" + section_link
        else:
            fname = "plot-main"
        fname = fname.replace("_", "-")
        if with_folder:
            return path.join(folder_name, fname)
        else:
            return fname

    if regex_plotsections is not None:
        do_plot = False
    else:
        do_plot = True

    if include_md:
        with open(path.join(folder_name, md_name), "w") as F_md:
            docdb_md_section(
                F_md,
                doc_db_root,
                section_history=[],
                heading_level=0,
                md_verbosity=None,
                plot_fname_func=plot_fname_func,
                plot_verbosity=plot_verbosity,
                regex_secname=regex_plotsections,
                do_plot=do_plot,
                plot_format=plot_format,
                use_ext=use_ext,
            )

    if include_plots:
        with asavefig.pool(workers=MP_workers):
            docdb_md_plots(
                doc_db_root,
                section_history=[],
                plot_verbosity=plot_verbosity,
                plot_fname_func=plot_fname_func,
                regex_secname=regex_plotsections,
                do_plot=do_plot,
                clear_plots=clear_plots,
                plot_format=plot_format,
                dpi=dpi,
            )

    if ipy_display:
        with open(path.join(folder_name, md_name), "r") as F_md:
            md_data = F_md.read()
        md_data = re.sub(r"\((.*)\.png\)", r"({0}/\1.png)".format(folder_name), md_data)
        try:
            from IPython.display import display_markdown
        except ImportError:
            print("Can't Import IPython.display")
        else:
            display_markdown(md_data, raw=True)

    if exporter is not None:
        exporter(
            md_name=md_name,
            folder=folder_name,
        )
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield.bunch.depbunch import DepBunch, Bunch

from ..roots_bin import roots_bin_type
from ..polynomials import poly_group


def abs_sq(x):
    return x.real ** 2 + x.imag ** 2


class DataFiltFitSplitBase(DepBunch):
    def __build__(
        self,
        _args=None,
        F_Hz=None,
        data=None,
        W=None,
        Fz_nyquist_Hz=None,
        Fp_nyquist_Hz=None,
        F_ref_Hz=None,
        delay_s=None,
        gain=None,
        **kwargs
    ):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")
        if F_Hz is None:
            raise RuntimeError("Must Specify F_Hz")
        if data is None:
            raise RuntimeError("Must Specify data")
        if W is None:
            W = 1
        if Fp_nyquist_Hz is None:
            raise RuntimeError("Must Specify Fp_nyquist_Hz")
        if Fz_nyquist_Hz is None:
            raise RuntimeError("Must Specify Fz_nyquist_Hz")
        if F_ref_Hz is None:
            F_ref_Hz = 0
        if delay_s is None:
            delay_s = 0
        if gain is None:
            gain = 1

        if data is not None and len(np.asarray(data).flatten()) == 0:
            data = None
        if W is not None and len(np.asarray(W).flatten()) == 0:
            W = None

        if data is not None:
            if W is None:
                W = 1
            F_Hz, data, W = np.broadcast_arrays(F_Hz, data, W)
        self.F_ref_Hz = F_ref_Hz
        self.Fz_nyquist_Hz = Fz_nyquist_Hz
        self.Fp_nyquist_Hz = Fp_nyquist_Hz
        self.F_Hz = F_Hz
        self.data = data
        self.W = W
        self.delay_s = delay_s
        self.gain = gain

        super(DataFiltFitSplitBase, self).__build__(**kwargs)

        @self.deco_generator
        def poly(self):
            return poly_group(self.Fp_nyquist_Hz)

        @self.deco_generator
        def poles_split(self):
            p_r, p_c, p_u, pol = roots_bin_type(
                self.poles, policy="strict", F_nyquist_Hz=self.Fp_nyquist_Hz
            )
            if len(p_u) > 0:
                raise RuntimeError(
                    "Got unorganized poles (missing conjugates). Not Prepared!"
                )
            return p_r, p_c

        @self.deco_generator
        def zeros_split(self):
            z_r, z_c, z_u, pol = roots_bin_type(
                self.zeros, policy="strict", F_nyquist_Hz=self.Fz_nyquist_Hz
            )
            if len(z_u) > 0:
                raise RuntimeError(
                    "Got unorganized zeros (missing conjugates). Not Prepared!"
                )
            return z_r, z_c

        @self.deco_generator
        def residuals(self):
            if self.data is None:
                raise RuntimeError("Can't generate residuals as data was not Specified")
            debias_reweight = 1 / (0.001 + self.W ** 2)
            retB = Bunch()
            retB.resP = self.W * (self.xfer_fit / self.data - 1)
            retB.resZ = self.W * (self.data / self.xfer_fit - 1)
            retB.total = np.sum(
                (abs(retB.resP) ** 2 + abs(retB.resZ * debias_reweight) ** 2)
                / (1 + debias_reweight)
            ) / (2 * len(self.data))
            return retB

        @self.deco_generator
        def F_max_Hz(self):
            return np.max(self.F_Hz)

        @self.deco_generator
        def Xp_grid_P(self):
            return np.exp(1j * np.pi * self.F_Hz / self.Fp_nyquist_Hz)

        @self.deco_generator
        def Xp_ref_P(self):
            return np.exp(1j * np.pi * self.F_ref_Hz / self.Fp_nyquist_Hz)

        @self.deco_generator
        def Xn_grid_P(self):
            return np.exp(-1j * np.pi * self.F_Hz / self.Fp_nyquist_Hz)

        @self.deco_generator
        def Xn_ref_P(self):
            return np.exp(-1j * np.pi * self.F_ref_Hz / self.Fp_nyquist_Hz)

        @self.deco_generator
        def Xp_grid_Z(self):
            return np.exp(1j * np.pi * self.F_Hz / self.Fz_nyquist_Hz)

        @self.deco_generator
        def Xp_ref_Z(self):
            return np.exp(1j * np.pi * self.F_ref_Hz / self.Fz_nyquist_Hz)

        @self.deco_generator
        def Xn_grid_Z(self):
            return np.exp(-1j * np.pi * self.F_Hz / self.Fz_nyquist_Hz)

        @self.deco_generator
        def Xn_ref_Z(self):
            return np.exp(-1j * np.pi * self.F_ref_Hz / self.Fz_nyquist_Hz)

    @property
    def ZPK(self):
        # cheap to compute, no need to autodelete
        return (self.zeros, self.poles, self.gain)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from ..codings_cmn import (
    CodingTypeZ,
    Ipi,
)


class CodingFBW(CodingTypeZ):
    N_parameters = 2
    p_F_Hz = 0
    p_BW = 0
    restabilize = False

    def update(self, F_Hz, BW):
        self.p_F_Hz = F_Hz
        self.p_BW = BW

    def reduce(self):
        return [
            self.p_F_Hz,
            self.p_BW,
        ]

    def update_roots(self, r1):
        amp = abs(r1)
        if self.restabilize:
            if amp > 1:
                print("stabilizing: ", r1)
                amp = 1 / amp
                self.lock_BW = True
        self.p_BW = (1 - amp) * self.sys.F_nyquist_Hz
        self.p_F_Hz = np.angle(r1) / np.pi * self.sys.F_nyquist_Hz
        return

    @property
    def F_Hz(self):
        return self.p_F_Hz

    @property
    def gain_effect(self):
        # always 1 since it is complex
        if self.p_BW >= 0:
            return 1
        else:
            return 1

    def transfer(self):
        amp = 1 - self.p_BW / self.sys.F_nyquist_Hz
        r = amp * np.cos(np.pi * self.p_F_Hz / self.sys.F_nyquist_Hz)
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        return (amp * amp) * Xnsq - 2 * Xn * r + 1

    def derivative(self):
        if self.disable:
            return []
        amp = 1 - self.p_BW / self.sys.F_nyquist_Hz
        F_nyquist_Hz = self.sys.F_nyquist_Hz
        Xn = self.sys.Xzn_grid
        Xnsq = self.sys.Xzn_grid_sq
        rcos = np.cos(np.pi * self.p_F_Hz / F_nyquist_Hz)
        rsin = np.sin(np.pi * self.p_F_Hz / F_nyquist_Hz)
        return [
            (2 * amp * np.pi / self.sys.F_nyquist_Hz) * (Xn * rsin),
            -((2 * amp) * Xnsq - Xn * (2 * rcos)) / self.sys.F_nyquist_Hz,
        ]

    def roots_c(self):
        # frequency, linear BW
        amp = 1 - self.p_BW / self.sys.F_nyquist_Hz
        return [amp * np.exp(Ipi * abs(self.p_F_Hz) / self.sys.F_nyquist_Hz)]



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import os.path as path

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import v2
from wield.iirrational.v2 import testing

from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
)


def test_simple1(request, browser, plotsections, plot_verbosity):
    data_name = "simple1"
    data = IIRrational_data(
        data_name,
        instance_num=1,
        set_num=1,
    )
    out = v2.data2filter(
        xfer=data.data,
        F_Hz=data.F_Hz,
        SNR=data.SNR,
        order_initial=10,
        delay_s=0,
        F_nyquist_Hz=None,
        poles=(-1, -1, -1, -1, -1),
        zeros=(-0.1, -0.1, -0.1),
        mode="fit",
        # hints = [
        #    testing.validate_plot_log(__file__, request),
        # ],
    )

    print(out.as_scipy_signal_ZPKsw())
    print(out.as_foton_ZPKsf())
    print(out.as_foton_str_ZPKsf())
    print(out.as_foton_ZPKsn())
    print(out.as_foton_str_ZPKsn())
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np


def fit_AAA(aid, order_hint=None):
    if order_hint is None:
        return fit_AAA_base(
            aid,
            order=aid.hint(
                "rational_AAA_fit_order",
                "rational_fit_order",
                "order_initial",
            ),
            order_max=aid.hint(
                "rational_AAA_fit_order_max",
                "rational_fit_order_max",
                "order_max",
            ),
            order_min=aid.hint(
                "rational_AAA_fit_order_min",
                "rational_fit_order_min",
                "order_min",
            ),
        )
    else:
        return fit_AAA_base(
            aid,
            order=aid.hint(
                order_hint,
                "rational_AAA_fit_order",
                "rational_fit_order",
                "order_initial",
            ),
            order_max=aid.hint(
                "rational_AAA_fit_order_max",
                "rational_fit_order_max",
                "order_max",
            ),
            order_min=aid.hint(
                "rational_AAA_fit_order_min",
                "rational_fit_order_min",
                "order_min",
            ),
        )


def fit_AAA_base(
    aid,
    order,
    order_max,
    order_min,
    account_size=True,
):
    if order == 0:
        return

    aid.log_progress(4, "AAA rational fit")
    # rat_fitter = v2.disc_sequence.rational_disc_fit(
    reldeg = aid.hint("relative_degree")
    if reldeg is None:
        reldeg_max = aid.hint("relative_degree_max")
        reldeg_min = aid.hint("relative_degree_min")
        if reldeg_min is None and reldeg_max is None:
            reldeg = None
        elif reldeg_min is None:
            reldeg = 0
        elif reldeg_max is None:
            reldeg = 0
        else:
            reldeg = int((reldeg_min + reldeg_max) // 2)

    factor_orders = aid.fitter_orders("factors")
    if reldeg is not None:
        diff_reldeg = reldeg - factor_orders.reldeg
    else:
        diff_reldeg = 0

    from wield.control.AAA import tfAAA
    if order is None:
        order = 20

    if order_max is None:
        order_max = order
    else:
        print(order, order_max)
        order_max = min(order, order_max)

    if account_size:
        factors_order = aid.fitter_orders().factors_maxzp
        # print("FACTORS ORDER", factors_order)
        order_max = order_max - factors_order
        order_max = max(order_max, 6)

    supports = np.linspace(aid.fitter.F_Hz[0], aid.fitter.F_Hz[-1], 5)
    rsum = np.cumsum(aid.fitter.residuals_sq)
    rsum = rsum / rsum[-1]
    sratios = np.linspace(0, 1, 8)[:-1]
    idx = np.searchsorted(rsum, sratios)
    supports = aid.fitter.F_Hz[idx]

    aaa = tfAAA(
        aid.fitter.F_Hz,
        aid.fitter.data_no_overlay,
        exact=False,
        res_tol=None,
        s_tol=None,
        w=aid.fitter.W,
        w_res=None,
        degree_max=order_max,
        nconv=None,
        nrel=10,
        rtype="log",
        lf_eager=True,
        supports=supports,
        minreal_cutoff=None,
        all_real=False
    )


    from wield.control import SISO

    #order = aaa.order
    #order_orig = order
    #while order > 1:
    #    aaa.choose(order)
    #    zeros = aaa.zeros
    #    poles = aaa.poles
    #    gain = aaa.gain
    #    select = poles.real > 0
    #    num_unstable = np.sum(select)
    #    
    #    aid.log_progress(4, "AAA order {}, num unstable: {} of {}, residuals {:.2f}".format(order, num_unstable, len(poles), aaa.fit_dict['res_rms']**2))
    #    if num_unstable == 0:
    #        break
    #    order -= 1
    #if order == 0:
    #    aaa.choose(order_orig)
    #    aid.log_progress(4, "AAA always unstable, using order {}".format(aaa.order))

    zeros = aaa.zeros
    poles = aaa.poles
    gain = aaa.gain

    from .. import fitters_ZPK
    fitter_bad = aid.fitter.regenerate(
        ZPKrep=aid.fitter.ZPKrep,
        coding_map=fitters_ZPK.codings_s.coding_maps.SOS,
        poles=np.asarray(poles),
        zeros=np.asarray(zeros),
        gain=gain,
        check_sign=True,
    )


    #from wield.utilities.mpl import mplfigB
    #axB = mplfigB(Nrows=2)
    #
    #h = fitter_bad.data_no_overlay
    #axB.ax0.loglog(fitter_bad.F_Hz, abs(h), marker='.', ls='')
    #axB.ax1.semilogx(fitter_bad.F_Hz, np.angle(h, deg=True), marker='.', ls='')

    #h = aaa(fitter_bad.F_Hz)
    #axB.ax0.loglog(fitter_bad.F_Hz, abs(h))
    #axB.ax1.semilogx(fitter_bad.F_Hz, np.angle(h, deg=True))


    #h = fitter_bad.xfer_fit_no_overlay
    #axB.ax0.loglog(fitter_bad.F_Hz, abs(h))
    #axB.ax1.semilogx(fitter_bad.F_Hz, np.angle(h, deg=True))

    # TODO, add debug_AAA hint
    # from .. import plots
    # axB = plots.plot_fitter_flag_residuals(fitter=aid.fitter, xscale='log')
    # axB.save("AAA_pre_{}.pdf".format(aid.N_update))
    # axB = plots.plot_fitter_flag_residuals(fitter=fitter_bad, xscale='log')
    # axB.save("AAA_{}.pdf".format(aid.N_update))

    with fitter_bad.with_codings_only([fitter_bad.gain_coding]):
        fitter_bad.optimize()
    fitter_bad.optimize()

    poles = fitter_bad.poles.fullplane
    select = poles.real > 0
    poles[select] = -poles[select].conjugate()
    gain = (-1)**np.sum(select) * gain

    fitter = aid.fitter.regenerate(
        ZPKrep=aid.fitter.ZPKrep,
        zeros=fitter_bad.zeros,
        poles=poles,
        gain=fitter_bad.gain,
    )
    fitter.optimize()
    assert(np.all(fitter.poles.fullplane.real < 0))

    #h = fitter_bad.xfer_fit_no_overlay
    #axB.ax0.loglog(fitter_bad.F_Hz, abs(h))
    #axB.ax1.semilogx(fitter_bad.F_Hz, np.angle(h, deg=True))

    #h = fitter.xfer_fit_no_overlay
    #axB.ax0.loglog(fitter.F_Hz, abs(h))
    #axB.ax1.semilogx(fitter.F_Hz, np.angle(h, deg=True))

    #for z in aaa.supports:
    #    axB.ax0.axvline(z, lw=0.5, ls='--', color='black')

    #axB.save("AAA_tests_{}.pdf".format(aid.N_update))

    #print("AAAlist: ", poles, zeros)

    aid.fitter_update(
        fitter,
        representative=False,
    )
    return




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield.bunch.depbunch import DepBunch, depB_property
from wield import declarative
from wield.bunch import Bunch


from .. import representations


class DataFiltFitBase(DepBunch):
    def __build__(
        self,
        _args=None,
        F_Hz=None,
        data=None,
        W=None,
        parent=None,
        codings=None,
        F_cutoff_Hz=None,
        **kwargs
    ):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")
        if parent is not None:
            if F_Hz is None:
                F_Hz = parent.F_Hz
            if data is None:
                data = parent.data
            if W is None:
                W = parent.W
            if codings is None:
                codings = parent.codings
            if F_cutoff_Hz is None:
                F_cutoff_Hz = parent.F_cutoff_Hz
        else:
            if F_Hz is None:
                raise RuntimeError("Must Specify F_Hz")
            if data is None:
                raise RuntimeError("Must Specify data")
            if W is None:
                W = 1

        if data is not None and len(np.asarray(data).flatten()) == 0:
            data = None
        if W is not None and len(np.asarray(W).flatten()) == 0:
            W = None

        if data is not None:
            if W is None:
                W = 1
            F_Hz, data, W = np.broadcast_arrays(F_Hz, data, W)
        self.F_Hz = F_Hz
        self.data = data
        self.W = W
        self.codings = codings

        if F_cutoff_Hz is not None:
            self.F_cutoff_Hz = F_cutoff_Hz

        super(DataFiltFitBase, self).__build__(**kwargs)

    @depB_property
    def residuals(self):
        if self.data is None:
            raise RuntimeError("Can't generate residuals as data was not Specified")
        debias_reweight = 1 / (0.001 + self.W ** 2)
        retB = Bunch()
        retB.resP = self.W * (self.xfer_fit / self.data - 1)
        retB.resZ = self.W * (self.data / self.xfer_fit - 1)
        retB.total = np.sum(
            (abs(retB.resP) ** 2 + abs(retB.resZ * debias_reweight) ** 2)
            / (1 + debias_reweight)
        ) / (2 * len(self.data))
        return retB

    @depB_property
    def F_max_Hz(self):
        return np.max(self.F_Hz)

    @depB_property
    def F_cutoff_Hz(self):
        return 1.00 * np.max(self.F_Hz)

    @depB_property
    def order(self):
        return max(
            len(self.poles) + len(self.poles_overlay),
            len(self.zeros) + len(self.zeros_overlay),
        )

    @depB_property
    def order_total(self):
        return (
            len(self.poles)
            + len(self.poles_overlay)
            + len(self.zeros)
            + len(self.zeros_overlay)
        )

    @depB_property
    def order_sos(self):
        return (
            max(
                len(self.poles) + len(self.poles_overlay),
                len(self.zeros) + len(self.zeros_overlay),
            )
            + 1
        ) // 2

    @depB_property
    def order_relative(self):
        return (len(self.zeros) + len(self.zeros_overlay)) - (
            len(self.poles) + len(self.poles_overlay)
        )

    @depB_property
    def ZPKrep(self):
        # cheap to compute, no need to autodelete
        self.codings_revision
        return representations.ZPKwData(
            data=self.data,
            W=self.W,
            F_Hz=self.F_Hz,
            zeros=self.zeros,
            poles=self.poles,
            zeros_overlay=self.zeros_overlay,
            poles_overlay=self.poles_overlay,
            gain=self.gain,
            delay_s=self.delay_s,
            F_nyquist_Hz=self.F_nyquist_Hz,
        )


class DataFiltFitZ(DataFiltFitBase):
    def __build__(
        self, _args=None, F_nyquist_Hz=declarative.NOARG, parent=None, **kwargs
    ):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")
        if parent is not None:
            if F_nyquist_Hz is declarative.NOARG:
                F_nyquist_Hz = parent.F_nyquist_Hz
        else:
            if F_nyquist_Hz is declarative.NOARG:
                raise RuntimeError("Must Specify F_nyquist_Hz")

        if F_nyquist_Hz is None:
            raise RuntimeError("This type cannot support F_nyquist_Hz = None")

        self.F_nyquist_Hz = F_nyquist_Hz

        super(DataFiltFitZ, self).__build__(parent=parent, **kwargs)

    @depB_property
    def Xzp_grid(self):
        return np.exp(1j * np.pi * self.F_Hz / self.F_nyquist_Hz)

    @depB_property
    def Xzp_grid_sq(self):
        return self.Xzp_grid * self.Xzp_grid

    @depB_property
    def Xzn_grid(self):
        return np.exp(-1j * np.pi * self.F_Hz / self.F_nyquist_Hz)

    @depB_property
    def Xzn_grid_sq(self):
        return self.Xzn_grid * self.Xzn_grid

    @depB_property
    def Xex_grid(self):
        return self.Xzp_grid


class DataFiltFitSf(DataFiltFitBase):
    def __build__(
        self, _args=None, F_nyquist_Hz=declarative.NOARG, parent=None, **kwargs
    ):
        if _args is not None:
            raise RuntimeError("Only keyword arguments allowed")
        if parent is not None:
            if F_nyquist_Hz is declarative.NOARG:
                F_nyquist_Hz = parent.F_nyquist_Hz
        else:
            if F_nyquist_Hz is declarative.NOARG:
                F_nyquist_Hz = None

        if F_nyquist_Hz is not None:
            raise RuntimeError("This type cannot support F_nyquist_Hz != None")

        self.F_nyquist_Hz = F_nyquist_Hz

        super(DataFiltFitSf, self).__build__(parent=parent, **kwargs)

    @depB_property
    def Xsf_grid(self):
        return 1j * self.F_Hz

    @depB_property
    def Xsf_grid_sq(self):
        return self.Xsf_grid * self.Xsf_grid

    @depB_property
    def Xex_grid(self):
        return self.Xsf_grid



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np

from . import roots_matching

conj_tol = 1e-12
real_tol = 1e-12
conj_tol = 1e-2
real_tol = 1e-4
conj_tol = 1e-6
real_tol = 1e-6


def are_conjugates(r1, r2):
    r2 = np.conjugate(r2)
    if abs(r1) < 0.8:
        return abs(r1 - r2) < conj_tol
    else:
        return abs((r1 / r2) - 1) < conj_tol


def is_real(r1):
    if abs(r1.real) < 0.8:
        return abs(np.imag(r1)) < real_tol
    else:
        return abs(np.imag(r1) / np.real(r1)) < real_tol


def roots_re_pair(r_r, r_c):
    roots = list(r_r)
    for r in r_c:
        roots.append(r)
        roots.append(r.conjugate())
    return np.asarray(roots)


def roots_bin_type(
    roots,
    policy="auto",
    F_nyquist_Hz=None,
    strict=True,
    simple_output=False,
    real_tol=1e-6,
    conj_tol=1e-4,
):
    """
    roots_r are the real roots
    roots_c are just the positive complex roots
    roots_u are any unsorted roots
    """
    assert policy in ["auto", "pos", "strict", "drop_pos", "drop_neg"]

    def are_same(r1, r2):
        if abs(r1) < 0.8:
            return abs(r1 - r2) < conj_tol
        else:
            return abs((r1 / r2) - 1) < conj_tol

    @np.vectorize
    def are_real(r1):
        if abs(r1.real) < 0.8:
            return abs(np.imag(r1)) < real_tol
        else:
            return abs(np.imag(r1) / np.real(r1)) < real_tol

    roots_r = []
    roots_u = []
    seen_cplx_pos = False
    seen_cplx_neg = False

    roots_u = []
    for root in roots:
        if are_real(root):
            roots_r.append(np.real(root))
        elif np.imag(root) > 0:
            seen_cplx_pos |= True
            roots_u.append(root)
        else:
            seen_cplx_neg |= True
            roots_u.append(root)

    if policy == "auto":
        if seen_cplx_neg and seen_cplx_pos:
            policy = "strict"
        else:
            policy = "pos"

    roots_c = []
    if policy == "strict":
        roots_u = np.asarray(roots_u)
        pos_select = roots_u.imag > 0
        roots_neg = roots_u[~pos_select]
        roots_pos = roots_u[pos_select]
        rB = roots_matching.nearest_unique_pairs(roots_pos, roots_neg.conjugate())
        roots_u = list(rB.l1_remain) + [r.conjugate() for r in rB.l2_remain]
        for r1, r2 in rB.r12_list:
            if not strict or are_same(r1, r2):
                # roots_c.append((r1 + r2) / 2)
                # TODO, this seems to work better, not clear why..
                roots_c.append(r1)
            else:
                roots_u.append(r1)
                roots_u.append(r2.conjugate())
    elif policy == "pos":
        for root in roots_u:
            if np.imag(root) > 0:
                roots_c.append(root)
            else:
                roots_c.append(root.conjugate())
            roots_u = []
    elif policy == "drop_neg":
        for root in roots_u:
            if np.imag(root) > 0:
                roots_c.append(root)
            roots_u = []
    elif policy == "drop_pos":
        for root in roots_u:
            if np.imag(root) < 0:
                roots_c.append(root.conjugate())
            roots_u = []

    if not simple_output:
        return (np.asarray(roots_r), np.asarray(roots_c), np.asarray(roots_u), policy)
    else:
        if roots_u:
            raise RuntimeError("Unbalanced or unpairable roots!")
        return roots_r, roots_c



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np

from .base import (
    CodingType,
    Ipi,
    # I2pi
)

# import scipy.linalg


class CodingnlFnlBW(CodingType):
    N_parameters = 2
    unstable = False
    p_nlF_Hz = 0
    p_nl_BW = 0
    lock_nlF_Hz = False
    lock_nl_BW = False
    hide_nlF_Hz = False
    hide_nl_BW = False
    min_BW_Hz = 0

    def setup(
        self,
        hide_all=None,
        lock_nlF_Hz=None,
        hide_nlF_Hz=None,
        lock_nl_BW=None,
        hide_nl_BW=None,
        hide_amplitude=None,
        disable=None,
    ):
        if hide_all:
            hide_nlF_Hz = True
            hide_nl_BW = True

        if hide_amplitude is not None:
            self.hide_nl_BW = hide_amplitude

        N_parameters = 2
        if lock_nlF_Hz is not None:
            self.lock_nlF_Hz = lock_nlF_Hz
        if hide_nlF_Hz is not None:
            self.hide_nlF_Hz = hide_nlF_Hz
        if self.hide_nlF_Hz:
            N_parameters -= 1

        if lock_nl_BW is not None:
            self.lock_nl_BW = lock_nl_BW
        if hide_nl_BW is not None:
            self.hide_nl_BW = hide_nl_BW
        if self.hide_nl_BW:
            N_parameters -= 1

        if disable is not None:
            self.disable = disable
            if disable:
                N_parameters = 0

        self.N_parameters = N_parameters

    def update(self, A=None, B=None):
        if self.disable:
            assert A is None and B is None
            return
        if self.hide_nl_BW:
            assert B is None
            if self.hide_nlF_Hz:
                assert A is None
            else:
                nlF_Hz = A
                if not self.lock_nlF_Hz:
                    self.p_nlF_Hz = nlF_Hz
        else:
            if self.hide_nlF_Hz:

                nl_BW = A
                if not self.lock_nl_BW:
                    self.p_nl_BW = nl_BW
            else:
                nlF_Hz, nl_BW = A, B
                if not self.lock_nlF_Hz:
                    self.p_nlF_Hz = nlF_Hz

                if not self.lock_nl_BW:
                    self.p_nl_BW = nl_BW

    def reduce(self):
        if self.disable:
            return []
        if self.hide_nl_BW:
            if self.hide_nlF_Hz:
                return []
            else:
                return [self.p_nlF_Hz]
        else:
            if self.hide_nlF_Hz:
                return [self.p_nl_BW]
            else:
                return [
                    self.p_nlF_Hz,
                    self.p_nl_BW,
                ]

    def option_set(self, minimum_BW_Hz=None, **kwargs):
        super(CodingnlFnlBW, self).option_set(**kwargs)
        if minimum_BW_Hz is not None:
            self.min_BW_Hz = minimum_BW_Hz
        # TODO, should modify p_nl_BW for the new min_BW_Hz
        return

    def update_roots(self, r1):
        """ """
        # TODO should honor min_BW_Hz
        F_Hz = r1.imag
        relF_Hz = F_Hz / (sys.F_cutoff_Hz)
        if relF_Hz >= 1:
            relF_Hz = 0.999
        elif relF_Hz <= -1:
            relF_Hz = 0.001
        self.p_nlF_Hz = relF_Hz / (1 - relF_Hz ** 2) ** 0.5

        if r1.real <= 0:
            self.unstable = False
            self.p_nl_BW = (-r1.real) ** 0.5
        else:
            self.unstable = True
            self.p_nl_BW = (r1.real) ** 0.5
        return

    @property
    def relF_Hz(self):
        val = self.p_nlF_Hz / (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        return val

    @property
    def BW_Hz(self):
        return self.min_BW_Hz + self.p_nl_BW ** 2

    def transfer(self, sys):
        # frequency, logarithmic amplitude
        if not self.unstable:
            r = -self.BW_Hz
        else:
            r = self.BW_Hz
        i = sys.F_cutoff_Hz * self.relF_Hz
        X = sys.Xsf_grid
        Xsq = sys.Xsf_grid_sq
        return (r * r + i * i) - 2 * X * r + Xsq

    def derivative(self, sys):
        if self.disable:
            return []
        # real/imaginary part of root
        if not self.unstable:
            r = -self.BW_Hz
            DrDp = -2 * self.p_nl_BW
        else:
            r = self.BW_Hz
            DrDp = 2 * self.p_nl_BW

        Fsqp5 = (1.0 + self.p_nlF_Hz ** 2) ** 0.5
        relF_Hz = self.p_nlF_Hz / Fsqp5
        DrelFHzDp = 1 / Fsqp5 - self.p_nlF_Hz ** 2 / (1.0 + self.p_nlF_Hz ** 2) ** 1.5
        i = sys.F_cutoff_Hz * relF_Hz

        X = sys.Xsf_grid
        if not self.hide_nlF_Hz:
            if not self.hide_nl_BW:
                return [
                    DrelFHzDp * sys.F_cutoff_Hz * 2 * i if not self.lock_nlF_Hz else 0,
                    2 * DrDp * r - 2 * X * DrDp if not self.lock_nl_BW else 0,
                ]
            else:
                return [
                    DrelFHzDp * sys.F_cutoff_Hz * 2 * i if not self.lock_nlF_Hz else 0,
                ]
        else:
            if not self.hide_nl_BW:
                return [
                    2 * DrDp * r - 2 * X * DrDp if not self.lock_nl_BW else 0,
                ]
            else:
                return []

    def roots_c(self):
        if not self.unstable:
            r = -self.BW_Hz
        else:
            r = self.BW_Hz
        i = sys.F_cutoff_Hz * self.relF_Hz
        # real/imaginary part of root
        return [r + 1j * i]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import numpy as np
import warnings
from .constraints import poly_constraints
from ..root_bunch import root_constraints, RBAlgorithms

log2 = np.log(2)

RBalgo = RBAlgorithms(
    strict=False,
    lax_line_tol=5e-1,
)


def coeff_canonicalization_gain(c):
    # print('CCG: ', c[-1])
    return c[-1]


def roots(c):
    return np.polynomial.chebyshev.chebroots(c)


def roots_lnG(c):
    lnG = np.log(c[-1])
    return np.polynomial.chebyshev.chebroots(c), lnG


def roots_rB(c, constraint):
    if poly_constraints.even_real <= constraint:
        assert np.all(c[::2].imag == 0)
    if poly_constraints.odd_real <= constraint:
        assert np.all(c[1::2].imag == 0)
    if poly_constraints.odd_imag <= constraint:
        assert np.all(c[1::2].real == 0)

    if poly_constraints.no_constraint == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(rvec, constraint=root_constraints.no_constraint)
    elif poly_constraints.eRoR == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(rvec, constraint=root_constraints.mirror_real)
    elif poly_constraints.eRoI == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(
            rvec, constraint=root_constraints.mirror_imag, allow_unknown=True
        )
        if len(rB.u) > 0:
            warnings.warn(
                "Unmirrored root in mirror_imag polynomial root finder (need to upgrade this algorithm)"
            )
            # HACK
            # clear any unknown
            rB.u = np.array([])
    elif poly_constraints.eRoZ == constraint:
        rvec = roots(c)
        rB = RBalgo.expect(rvec, constraint=root_constraints.mirror_quad)
    return rB


def fromroots_lnG(roots):
    lnG = log2 * (len(roots) - 1)
    c = np.polynomial.chebyshev.chebfromroots(roots)
    cG = coeff_canonicalization_gain(c)
    c /= cG
    lnG += np.log(cG)
    return c, lnG


def val_lnG(X, c, lnG=0):
    # the LOG2 is because the last coefficient is assumed to be scaled to one (as is done in fromroots)
    lnG_out = -(len(c) - 2) * log2
    return np.polynomial.chebyshev.chebval(X, c), lnG + lnG_out


def vander_lnG(X, N, lnG=0):
    # the LOG2 is because the last coefficient is assumed to be scaled to one (as is done in fromroots)
    lnG_out = -(N - 1) * log2
    return np.polynomial.chebyshev.chebvander(X, N), lnG + lnG_out


def companion(c):
    return np.polynomial.chebyshev.chebcompanion(c)



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import numpy as np
import os.path as path
import matplotlib as mpl
import matplotlib.pyplot as plt


def simple1_interactive(browser, plotsections, plot_verbosity):

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.plot(np.linspace(1, 10, 100))

    plt.ioff()
    plt.show()
    return


if __name__ == "__main__":
    test_simple1(False, None, 10)



# -*- coding: utf-8 -*-
"""
"""
import pytest

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import testing

from wield.iirrational.fitters_ZPK import ZPKrep2MRF
from wield.iirrational.fitters_ZPK import codings_s

from utilities import check_residuals


@pytest.mark.parametrize("coding_map", sorted(codings_s.coding_maps.keys()))
def test_NL_fitter1(coding_map):
    """
    tests when the fitter is bad as it is trying to fit on the wrong side
    """
    data = IIRrational_data("testZPK4x")
    print(data.rep_s.zeros)
    data.rep_s.zeros = 0.99 * data.rep_s.zeros
    data.rep_s.poles = 0.99 * data.rep_s.poles
    print(data.rep_s.zeros)
    fitterMRF = ZPKrep2MRF(
        data.rep_s.ZPKrep,
        coding_map=codings_s.coding_maps[coding_map],
    )
    testing.assert_almost_equal(data.rep_s.xfer_fit / fitterMRF.xfer_fit, 1, 4)

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)


@pytest.mark.parametrize("coding_map", sorted(codings_s.coding_maps.keys()))
def test_NL_fitter2(coding_map):
    data = IIRrational_data("testZPK4x")
    print("Z:", data.rep_s.zeros)
    print("P:", data.rep_s.poles)
    data.rep_s.zeros = 1.01 * data.rep_s.zeros
    data.rep_s.poles = 0.99 * data.rep_s.poles
    fitterMRF = ZPKrep2MRF(
        data.rep_s.ZPKrep,
        coding_map=codings_s.coding_maps[coding_map],
    )

    testing.assert_almost_equal(data.rep_s.xfer_fit / fitterMRF.xfer_fit, 1, 4)

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    assert fitterMRF.residuals_average < 10

    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)


@pytest.mark.parametrize("coding_map", sorted(codings_s.coding_maps.keys()))
def test_NL_fitter1_DL(coding_map):
    """
    tests when the fitter is bad as it is trying to fit on the wrong side
    """
    # set number 1 is logarithmic data
    data = IIRrational_data("testZPK4xHQ", set_num=1)
    data.rep_s.zeros = 0.99 * data.rep_s.zeros
    data.rep_s.poles = 0.99 * data.rep_s.poles
    print(data.rep_s.zeros)
    fitterMRF = ZPKrep2MRF(
        data.rep_s.ZPKrep,
        coding_map=codings_s.coding_maps[coding_map],
        distance_limit_auto=1,
    )
    testing.assert_almost_equal(data.rep_s.xfer_fit / fitterMRF.xfer_fit, 1, 4)
    fitterMRF.distance_limit_auto = 1

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)


@pytest.mark.parametrize("coding_map", sorted(codings_s.coding_maps.keys()))
def test_NL_fitter1_DL2(coding_map):
    """
    tests when the fitter is bad as it is trying to fit on the wrong side
    """
    # set number 1 is logarithmic data
    data = IIRrational_data("testZPK4xHQ", set_num=1)
    data.rep_s.zeros = 0.99 * data.rep_s.zeros
    data.rep_s.poles = 0.99 * data.rep_s.poles
    print(data.rep_s.zeros)
    fitterMRF = ZPKrep2MRF(
        data.rep_s.ZPKrep,
        coding_map=codings_s.coding_maps[coding_map],
        distance_limit_auto=1,
    )
    testing.assert_almost_equal(data.rep_s.xfer_fit / fitterMRF.xfer_fit, 1, 4)
    fitterMRF.distance_limit_auto = 2

    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    opt = fitterMRF.optimize()
    print(fitterMRF.residuals_average)
    check_residuals(fitterMRF)
    # axB = plot_fitter_flag(fitterMRF, with_error=False, plot_zp = True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from wield.bunch.depbunch import depB_property, NOARG

from .. import representations
from ..representations.polynomials import standard

from .rational_algorithms import PolyFilterAlgorithms
from .rational_attributes import PolyFit
from .rational_bases import ZFilterBase


class RootConstraints(representations.RootConstraints):
    branch_separated = frozenset("branch_separated")


root_constraints = RootConstraints()


# must be in this order for the ZPKz to be set in the correct order
class RationalDiscFilter(ZFilterBase, PolyFit, PolyFilterAlgorithms):
    phase_missing = False
    poly = standard
    stablize_pos_method = "ignore"
    stablize_neg_method = "ignore"

    @depB_property
    def zeros_phase_adjust(self, val=NOARG):
        if val is NOARG:
            return None
        else:
            return val

    @depB_property
    def poles_phase_adjust(self, val=NOARG):
        if val is NOARG:
            return None
        else:
            return val

    @depB_property
    def zeros_phasing(self):
        if self.zeros_phase_adjust is None:
            return self.Xn_grid ** self.nzeros
        elif self.zeros_phase_adjust < 0:
            return self.Xn_grid ** (self.nzeros + self.zeros_phase_adjust)
        else:
            return self.Xn_grid ** (self.zeros_phase_adjust)
        return

    @depB_property
    def poles_phasing(self):
        if self.poles_phase_adjust is None:
            return self.Xn_grid ** self.npoles
        elif self.poles_phase_adjust < 0:
            return self.Xn_grid ** (self.npoles + self.poles_phase_adjust)
        else:
            return self.Xn_grid ** (self.poles_phase_adjust)
        return

    def _phasing_roots(self, roots):
        return self.Xn_grid ** len(roots)

    def _phasing_vec(self, vec):
        return self.Xn_grid ** (len(vec) - 1)

    def _phasing_afit(self):
        return self.poles_phasing

    def _phasing_bfit(self):
        return self.zeros_phasing

    def root_stabilize(
        self,
        rB,
        method=None,
        real_neg=None,
        real_pos=None,
    ):
        if method is None:
            return rB

        if real_neg is None:
            real_neg = self.stablize_neg_method
        if real_pos is None:
            real_pos = self.stablize_pos_method

        r_r = np.copy(rB.r)
        r_c = np.copy(rB.c)

        if real_neg == "use":
            select_r = r_r < -1
        elif real_neg == "ignore":
            select_r = np.zeros(r_r.shape, dtype=bool)
        else:
            raise RuntimeError("Bad Argument")

        if real_pos == "use":
            select_r = r_r > 1 | select_r
        elif real_pos == "ignore":
            pass
        else:
            raise RuntimeError("Bad Argument")

        select_c = abs(r_c) > 1
        if method == "flip":
            r_r[select_r] = 1 / r_r[select_r]
            r_c[select_c] = 1 / r_c[select_c].conjugate()
        elif method == "remove":
            r_r = r_r[~select_r]
            r_c = r_c[~select_c]
        else:
            raise RuntimeError("Bad Argument")

        rB = representations.RootBunch(
            constraint=self.root_constraint,
            r=r_r,
            c=r_c,
        )
        return rB

    def clear_unstable_poles(self):
        new_poles = []
        for r in self.poles:
            if abs(r) < 1:
                new_poles.append(r)
        self.poles = new_poles

    def clear_bigdelay_zeros(self):
        new_zeros = []
        for r in self.zeros:
            if abs(r) < 1:
                new_zeros.append(r)
        self.zeros = new_zeros

    def phi_rep_Snative(self, rB):
        """
        Maps the internal representation to Snative, the output RootBunch can
        have a specialty constraint added.
        """
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        # TODO
        # ok to modify?
        rB.constraint = rB.constraint | root_constraints.branch_separated
        select_pos = rB.r > 0
        rB.nyquist_branch = (-rB.r[~select_pos] - 1) * self.F_nyquist_Hz
        rB.r = (rB.r[select_pos] - 1) * self.F_nyquist_Hz
        real = (abs(rB.c) - 1) * self.F_nyquist_Hz
        imag = np.angle(rB.c) / np.pi * self.F_nyquist_Hz
        rB.c = real + 1j * imag
        return rB

    def phi_Snative_rep(self, rB):
        """ """
        assert root_constraints.mirror_real == (
            rB.constraint - root_constraints.branch_separated
        )
        rB.c = (1 + rB.c.real / self.F_nyquist_Hz) * np.exp(
            1j * rB.c.imag * np.pi / self.F_nyquist_Hz
        )

        if root_constraints.branch_separated <= rB.constraint:
            rB.r = np.concatenate(
                [
                    (rB.r / self.F_nyquist_Hz + 1),
                    -(rB.nyquist_branch / self.F_nyquist_Hz + 1),
                ]
            )
        else:
            rB.r = rB.r / self.F_nyquist_Hz + 1

        rB.constraint = rB.constraint - root_constraints.branch_separated
        rB = self.RBalgo.expect(rB, self.RBalgo.root_constraints.mirror_real)
        return rB



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch

from .base import (
    ArgumentError,
    grab_kwargs,
    grab_kwarg_hints,
    kw_ZPKrep_build,
    check_remaining_arguments,
    transfer_kw,
)
from . import logging
from . import standardargs
from . import adjustments
from . import ranges

kw_hints = Bunch()
kw_hints.update(logging.kw_hints)
kw_hints.update(standardargs.kw_hints)
kw_hints.update(adjustments.kw_hints)
kw_hints.update(ranges.kw_hints)

__all__ = [
    ArgumentError,
    grab_kwargs,
    grab_kwarg_hints,
    kw_ZPKrep_build,
    kw_hints,
    logging,
    standardargs,
    adjustments,
    ranges,
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

# import numpy as np

from ..codings_cmn import (
    CodingType,
)


class CodingRealBW(CodingType):
    N_parameters = 1
    unstable = False
    p_BW_Hz = 0
    restabilize = False

    def update(self, BW):
        self.p_BW_Hz = BW

    def reduce(self):
        return [self.p_BW_Hz]

    def update_roots(self, r1):
        assert r1.imag == 0
        if self.restabilize:
            if r1.real > 0:
                r1 = -r1
        self.p_BW_Hz = -r1.real
        return

    @property
    def F_Hz(self):
        return 0

    @property
    def BW(self):
        return self.p_BW_Hz

    @property
    def gain_effect(self):
        if self.p_BW_Hz < 0:
            return -1
        else:
            return 1

    def transfer(self):
        # real, linear BW
        r = -self.p_BW_Hz
        return self.sys.Xsf_grid - r

    def derivative(self):
        return [1]

    def roots_r(self):
        r = -self.p_BW_Hz
        return [r]



# -*- coding: utf-8 -*-
"""
This code originally from module scipy.optimize.optimize
modified slightly to reduce dependencies.

SciPy license
Copyright © 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright © 2003-2013 SciPy Developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer. Redistributions in binary
    form must reproduce the above copyright notice, this list of conditions and
    the following disclaimer in the documentation and/or other materials
    provided with the distribution. Neither the name of Enthought nor the names
    of the SciPy Developers may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Trust Region Reflective algorithm for least-squares optimization.

The algorithm is based on ideas from paper [STIR]_. The main idea is to
account for presence of the bounds by appropriate scaling of the variables (or
equivalently changing a trust-region shape). Let's introduce a vector v:

           | ub[i] - x[i], if g[i] < 0 and ub[i] < np.inf
    v[i] = | x[i] - lb[i], if g[i] > 0 and lb[i] > -np.inf
           | 1,           otherwise

where g is the gradient of a cost function and lb, ub are the bounds. Its
components are distances to the bounds at which the anti-gradient points (if
this distance is finite). Define a scaling matrix D = diag(v**0.5).
First-order optimality conditions can be stated as

    D^2 g(x) = 0.

Meaning that components of the gradient should be zero for strictly interior
variables, and components must point inside the feasible region for variables
on the bound.

Now consider this system of equations as a new optimization problem. If the
point x is strictly interior (not on the bound) then the left-hand side is
differentiable and the Newton step for it satisfies

    (D^2 H + diag(g) Jv) p = -D^2 g

where H is the Hessian matrix (or its J^T J approximation in least squares),
Jv is the Jacobian matrix of v with components -1, 1 or 0, such that all
elements of matrix C = diag(g) Jv are non-negative. Introduce the change
of the variables x = D x_h (_h would be "hat" in LaTeX). In the new variables
we have a Newton step satisfying

    B_h p_h = -g_h,

where B_h = D H D + C, g_h = D g. In least squares B_h = J_h^T J_h, where
J_h = J D. Note that J_h and g_h are proper Jacobian and gradient with respect
to "hat" variables. To guarantee global convergence we formulate a
trust-region problem based on the Newton step in the new variables:

    0.5 * p_h^T B_h p + g_h^T p_h -> min, ||p_h|| <= Delta

In the original space B = H + D^{-1} C D^{-1}, and the equivalent trust-region
problem is

    0.5 * p^T B p + g^T p -> min, ||D^{-1} p|| <= Delta

Here the meaning of the matrix D becomes more clear: it alters the shape
of a trust-region, such that large steps towards the bounds are not allowed.
In the implementation the trust-region problem is solved in "hat" space,
but handling of the bounds is done in the original space (see below and read
the code).

The introduction of the matrix D doesn't allow to ignore bounds, the algorithm
must keep iterates strictly feasible (to satisfy aforementioned
differentiability), the parameter theta controls step back from the boundary
(see the code for details).

The algorithm does another important trick. If the trust-region solution
doesn't fit into the bounds, then a reflected (from a firstly encountered
bound) search direction is considered. For motivation and analysis refer to
[STIR]_ paper (and other papers of the authors). In practice it doesn't need
a lot of justifications, the algorithm simply chooses the best step among
three: a constrained trust-region step, a reflected step and a constrained
Cauchy step (a minimizer along -g_h in "hat" space, or -D^2 g in the original
space).

Another feature is that a trust-region radius control strategy is modified to
account for appearance of the diagonal C matrix (called diag_h in the code).

Note, that all described peculiarities are completely gone as we consider
problems without bounds (the algorithm becomes a standard trust-region type
algorithm very similar to ones implemented in MINPACK).

The implementation supports two methods of solving the trust-region problem.
The first, called 'exact', applies SVD on Jacobian and then solves the problem
very accurately using the algorithm described in [JJMore]_. It is not
applicable to large problem. The second, called 'lsmr', uses the 2-D subspace
approach (sometimes called "indefinite dogleg"), where the problem is solved
in a subspace spanned by the gradient and the approximate Gauss-Newton step
found by ``scipy.sparse.linalg.lsmr``. A 2-D trust-region problem is
reformulated as a 4-th order algebraic equation and solved very accurately by
``numpy.roots``. The subspace approach allows to solve very large problems
(up to couple of millions of residuals on a regular PC), provided the Jacobian
matrix is sufficiently sparse.

References
----------
.. [STIR] Branch, M.A., T.F. Coleman, and Y. Li, "A Subspace, Interior,
      and Conjugate Gradient Method for Large-Scale Bound-Constrained
      Minimization Problems," SIAM Journal on Scientific Computing,
      Vol. 21, Number 1, pp 1-23, 1999.
.. [JJMore] More, J. J., "The Levenberg-Marquardt Algorithm: Implementation
    and Theory," Numerical Analysis, ed. G. A. Watson, Lecture
"""

import numpy as np
import numpy as numpy
from wield import declarative

from numpy.linalg import norm
from scipy.linalg import svd

from .common import (
    solve_lsq_trust_region,
    solve_trust_region_2d,
    minimize_quadratic_1d,
    build_quadratic_1d,
    right_multiplied_operator,
    evaluate_quadratic,
    compute_grad,
    compute_jac_scale,
    check_termination,
    update_tr_radius,
    scale_for_robust_loss_function,
    print_header_nonlinear,
    print_iteration_nonlinear,
)

try:
    from scipy.optimize import OptimizeResult
except ImportError:
    OptimizeResult = Bunch


def trf(fun, jac, x0, f0, J0, ftol, xtol, gtol, max_nfev, x_scale, verbose):
    x = x0.copy()

    f = f0
    f_true = f.copy()
    nfev = 1

    J = J0
    njev = 1
    m, n = J.shape

    cost = 0.5 * np.dot(f, f)

    g = compute_grad(J, f)

    jac_scale = isinstance(x_scale, str) and x_scale == "jac"
    if jac_scale:
        scale, scale_inv = compute_jac_scale(J)
    else:
        scale, scale_inv = x_scale, 1 / x_scale

    Delta = norm(x0 * scale_inv)
    if Delta == 0:
        Delta = 1.0

    if max_nfev is None:
        max_nfev = x0.size * 100

    alpha = 0.0  # "Levenberg-Marquardt" parameter

    termination_status = None
    iteration = 0
    step_norm = None
    actual_reduction = None

    if verbose == 2:
        print_header_nonlinear()

    while True:
        g_norm = norm(g, ord=np.inf)
        if g_norm < gtol:
            termination_status = 1

        if verbose == 2:
            print_iteration_nonlinear(
                iteration, nfev, cost, actual_reduction, step_norm, g_norm
            )

        if termination_status is not None or nfev == max_nfev:
            break

        d = scale
        g_h = d * g

        J_h = J * d
        U, s, V = svd(J_h, full_matrices=False)
        V = V.T
        uf = U.T.dot(f)

        actual_reduction = -1
        while actual_reduction <= 0 and nfev < max_nfev:
            step_h, alpha, n_iter = solve_lsq_trust_region(
                n, m, uf, s, V, Delta, initial_alpha=alpha
            )

            predicted_reduction = -evaluate_quadratic(J_h, g_h, step_h)
            step = d * step_h
            x_new = x + step
            f_new = fun(x_new)
            nfev += 1

            step_h_norm = norm(step_h)

            if not np.all(np.isfinite(f_new)):
                Delta = 0.25 * step_h_norm
                continue

            # Usual trust-region step quality estimation.
            cost_new = 0.5 * np.dot(f_new, f_new)
            actual_reduction = cost - cost_new

            Delta_new, ratio = update_tr_radius(
                Delta,
                actual_reduction,
                predicted_reduction,
                step_h_norm,
                step_h_norm > 0.95 * Delta,
            )
            alpha *= Delta / Delta_new
            Delta = Delta_new

            step_norm = norm(step)
            termination_status = check_termination(
                actual_reduction, cost, step_norm, norm(x), ratio, ftol, xtol
            )

            if termination_status is not None:
                break

        if actual_reduction > 0:
            x = x_new

            f = f_new
            f_true = f.copy()

            cost = cost_new

            J = jac(x, f)
            njev += 1

            g = compute_grad(J, f)

            if jac_scale:
                scale, scale_inv = compute_jac_scale(J, scale_inv)
        else:
            step_norm = 0
            actual_reduction = 0

        iteration += 1

    if termination_status is None:
        termination_status = 0

    active_mask = np.zeros_like(x)
    return OptimizeResult(
        x=x,
        cost=cost,
        fun=f_true,
        jac=J,
        grad=g,
        optimality=g_norm,
        active_mask=active_mask,
        nfev=nfev,
        njev=njev,
        status=termination_status,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from .data2filter import data2filter

# from .disc_sequence import rational_disc_fit
# from .disc_sequence_mag import rational_disc_fit_mag
# from .fit_aid import FitAid
from ..data2testcase import data2testcase
from . import hintsets
from .statespace import ss2filter, ss2xfer, ss2zpk

__all__ = [
    "data2filter",
    "hintsets,"
    "data2testcase",
    "ss2filter,"
    "ss2xfer",
    "ss2zpk,"
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


from ...fitters_rational import ChebychevFilter
from ...utilities import ensure_aid


def reldeg_split(relative_degree):
    if relative_degree is None or relative_degree == 0:
        reln = 0
        reld = 0
    elif relative_degree > 0:
        reln = relative_degree
        reld = 0
    else:
        reln = 0
        reld = -relative_degree
    return reln, reld


def rational_cheby_fit(
    ZPKrep,
    order=None,
    order_max=None,
    order_min=20,
    relative_degree=None,
    aid=None,
):
    aid = ensure_aid(aid)

    if ZPKrep.F_nyquist_Hz is not None:
        raise RuntimeError("v2 only supports S domain for now")

    if order is not None:
        order = int(order)
    order_max = int(order_max)
    order_min = int(order_min)

    N_first = order_min
    N_final = int(min(len(ZPKrep.F_Hz) // 10, order_max))

    if order is not None:
        ret = cheby_single(
            ZPKrep,
            relative_degree=relative_degree,
            order=order,
            aid=aid,
        )
        return ret

    # otherwise, scan through
    N_current = N_first

    fitter_last = cheby_single(
        ZPKrep,
        order=N_current,
        relative_degree=relative_degree,
        aid=aid,
    )
    restot_last = fitter_last.residuals_average

    def fitter_ord(fitter):
        return max(len(fitter.zeros), len(fitter.poles))

    with aid.log_heading("Rational Fit Order Scanning"):
        while True:
            if N_current == N_final:
                fitter_use = fitter_last
                break

            N_current = N_current * 2
            if N_current > N_final:
                N_current = N_final

            fitter = cheby_single(
                ZPKrep,
                order=N_current,
                relative_degree=relative_degree,
                aid=aid,
            )
            restot = fitter.residuals_average
            fitter_red = fitter.copy()
            fitter_red.matched_pairs_clear(Q_rank_cutoff=0.2)
            restot_red = fitter_red.residuals_average

            if restot_last < restot:
                fitter_use = fitter_last
                aid.log_debug(8, "Using last (direct)!", fitter_last.order)
                break

            if restot_last < restot_red:
                fitter_use = fitter_last
                aid.log_debug(8, "Using last (reduced)!", fitter_last.order)
                break

            if restot_last < 1.10 * restot:
                ord_red = fitter_red.order
                ord_last = fitter_last.order
                if ord_red < ord_last:
                    fitter_use = fitter_red
                else:
                    fitter_use = fitter_last
                aid.log_debug(8, "Using current")
                break
            # else continue the loop
            fitter_last = fitter
            restot_last = restot

    return fitter_use


def cheby_single(
    ZPKrep,
    order=20,
    relative_degree=0,
    aid=None,
):
    aid = ensure_aid(aid)

    if ZPKrep.F_nyquist_Hz is not None:
        raise RuntimeError("F_nyquist_Hz must be None (ZPKrep must be in S domain)")

    if relative_degree is None:
        relative_degree = 0

    reln, reld = reldeg_split(relative_degree)
    fitter = ChebychevFilter(
        ZPKrep=ZPKrep,
        nzeros=(order // 2) + reln,
        npoles=(order // 2) + reld,
    )

    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.nzeros = order + reln
    fitter.npoles = order + reld
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.5)
    fitter.stabilize = True
    # fitter.npoles = len(fitter.poles)
    # fitter.nzeros = len(fitter.zeros)
    ##fitter.stabilize = True
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_SVD()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.matched_pairs_clear(Q_rank_cutoff=0.5)
    # fitter.stabilize = False
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    fitter.fit_poles()
    fitter.fit_zeros()
    return fitter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import os
from msurrogate import SurrogateApp

from . import v1
from . import testing

from ..plots import IIRPlots

plots = IIRPlots(
    show_immediate=True,
    show_saved=False,
    force_save=False,
)

if __name__ == "__main__":
    app = SurrogateApp()
    # app.option('serializer-preferred', 'dill')
    app.option("serializer-preferred", "pickle")
    metaD = app.meta_daemon_setup()
    metaD.register(v1, "v1")
    metaD.register(v1, "v2")
    metaD.register(testing, "testing")
    metaD.register(plots, "plots")
    app.run_loop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""


import numpy as np
from ... import TFmath

Ipi = np.pi * 1j
I2pi = np.pi * 2j


class BranchCutAmbiguity(Exception):
    pass


class EmptyCopy(object):
    pass


class CodingType(object):
    disable = False
    coding_id = None

    gain_effect = 1

    def __init__(self, sys):
        # preserved through deepcopy so that codings_ignore set can be maintained
        self.coding_id = id(self)
        self.sys = sys

    def clone(self, sys):
        new = EmptyCopy()
        d = dict(self.__dict__)
        d["sys"] = sys
        new.__dict__.update(d)
        new.__class__ = self.__class__
        return new

    def roots(self):
        rc = self.roots_c()
        return self.roots_r() + rc + [r.conjugate() for r in rc]

    def roots_r(self):
        return []

    def roots_c(self):
        return []

    def roots_Sf(self):
        return self.roots()

    def roots_r_Sf(self):
        return self.roots_r()

    def roots_c_Sf(self):
        return self.roots_c()

    def update_roots_Sf(self, *rs):
        return self.update_roots(*rs)

    def option_set(self, **kwargs):
        return

    def transfer_abs_sq(self):
        # real/imaginary part of root
        return TFmath.abs_sq(self.transfer())

    def derivative_wtrans(self):
        return self.transfer(), self.derivative()

    def derivative_abs_sq_wtrans(self):
        xfer = self.transfer()
        jac = self.derivative()
        jac_abs_sq = []
        for der in jac:
            jac_abs_sq.append(2 * (der.real * xfer.real + der.imag * xfer.imag))
        return TFmath.abs_sq(xfer), jac_abs_sq

    @property
    def derivative_deadzoned(self):
        return False


class CodingTypeZ(object):
    disable = False
    coding_id = None

    gain_effect = 1

    def __init__(self, sys):
        # preserved through deepcopy so that codings_ignore set can be maintained
        self.coding_id = id(self)
        self.sys = sys

    def clone(self, sys):
        new = EmptyCopy()
        d = dict(self.__dict__)
        d["sys"] = sys
        new.__dict__.update(d)
        new.__class__ = self.__class__
        return new

    def roots(self):
        rc = self.roots_c()
        return self.roots_r() + rc + [r.conjugate() for r in rc]

    def roots_r(self):
        return []

    def roots_c(self):
        return []

    def roots_Sf(self):
        rs = []
        for r in self.roots():
            if r.imag == 0:
                if r.real > 0:
                    r_Sf = (r.real - 1) * self.sys.F_nyquist_Hz
                else:
                    raise BranchCutAmbiguity()
            else:
                F_Hz = np.angle(r) / np.pi * self.sys.F_nyquist_Hz
                amp = abs(r)
                BW = (amp - 1) * self.sys.F_nyquist_Hz
                r_Sf = BW + 1j * F_Hz
            rs.append(r_Sf)
        return rs

    def roots_r_Sf(self):
        rs = []
        for r in self.roots_r():
            if r.real > 0:
                r_Sf = (r.real - 1) * self.sys.F_nyquist_Hz
            else:
                raise BranchCutAmbiguity()
            rs.append(r_Sf)
        return rs

    def roots_c_Sf(self):
        rs = []
        for r in self.roots_c():
            F_Hz = np.angle(r) / np.pi * self.sys.F_nyquist_Hz
            amp = abs(r)
            BW = (amp - 1) * self.sys.F_nyquist_Hz
            r_Sf = BW + 1j * F_Hz
            rs.append(r_Sf)
        return rs

    def update_roots_Sf(self, *rs):
        rZs = []
        for r in rs:
            F_Hz = r.imag
            if F_Hz > self.sys.F_nyquist_Hz:
                raise BranchCutAmbiguity()
            amp = 1 + r.real / self.sys.F_nyquist_Hz
            rZ = amp * np.exp(F_Hz / self.sys.F_nyquist_Hz * np.pi * 1j)
            rZs.append(rZ)

        return self.update_roots(*rZs)

    def option_set(self, **kwargs):
        return

    def transfer_abs_sq(self):
        # real/imaginary part of root
        return TFmath.abs_sq(self.transfer())

    def derivative_wtrans(self):
        return self.transfer(), self.derivative()

    def derivative_abs_sq_wtrans(self):
        xfer = self.transfer()
        jac = self.derivative()
        jac_abs_sq = []
        for der in jac:
            jac_abs_sq.append(2 * (der.real * xfer.real + der.imag * xfer.imag))
        return TFmath.abs_sq(xfer), jac_abs_sq

    @property
    def derivative_deadzoned(self):
        return False



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import os.path as path

from wield.iirrational.testing import IIRrational_data
from wield.iirrational import v2
from wield.iirrational.v2 import testing

from wield.iirrational.testing.utilities import (
    sign_validate_and_plot_hint,
)


def test_simple1(request, browser, plotsections, plot_verbosity):
    data_name = "simple1"

    dset = IIRrational_data(
        data_name,
        instance_num=1,
        set_num=1,
    )
    out = v2.data2filter(
        F_Hz=dset.F_Hz,
        SNR=dset.SNR,
        data=dset.data,
        order_initial=10,
        delay_s=0,
        F_nyquist_Hz=None,
        hints=[
            testing.validate_plot_log(__file__, request),
        ],
    )

    folder = path.join(path.split(__file__)[0], "plots")
    axB = out.investigate_order_plot()
    axB.save(path.join(folder, data_name))
    return


def test_simple1_fit(request, browser, plotsections, plot_verbosity):
    data_name = "simple1"
    dset = IIRrational_data(
        data_name,
        instance_num=1,
        set_num=1,
    )
    out = v2.data2filter(
        F_Hz=dset.F_Hz,
        SNR=dset.SNR,
        data=dset.data,
        ZPK=dset.bestfit_ZPK_s,
        order_initial=10,
        delay_s=0,
        F_nyquist_Hz=None,
        mode="fit",
        hints=[
            testing.validate_plot_log(__file__, request),
        ],
    )

    with testing.plot_on_assert(__file__, request, out.fitter, plot_anyway=False):
        pass
    print(out.as_foton_str_ZPKsf())
    return



# -*- coding: utf-8 -*-
"""
"""
from __future__ import division, print_function, unicode_literals
import pytest

import os.path as path

# from wield.iirrational import fitters
from wield.iirrational.fitters_rational.polynomials import standard
import numpy as np

try:
    from IIRrational_test_data import IIRrational_data_dev
except ImportError:
    module_import_skip = True
else:
    module_import_skip = False


def test_std_4roots_1scale():
    X_scale = 1
    roots = [-0.8 * X_scale, 0 + 0.5j * X_scale, 0 - 0.5j * X_scale, -0.5 * X_scale]
    c, lnG = standard.fromroots_lnG(roots, X_scale=X_scale)

    F = np.linspace(0, X_scale, X_scale)
    X = 1j * F

    y_main, y_main_lnG = standard.valfromroots_lnG(X, roots, X_scale=X_scale)
    y_poly, y_poly_lnG = standard.val_lnG(X, c, X_scale=X_scale, lnG=lnG)
    V, VlnG = standard.vander_lnG(X, len(c) - 1, X_scale=X_scale, lnG=lnG)
    y_vander = np.dot(V, c) * np.exp(VlnG)

    np.testing.assert_allclose(y_vander, y_poly * np.exp(y_poly_lnG))
    np.testing.assert_allclose(y_main * np.exp(y_main_lnG), y_vander)
    return


# @pytest.mark.skipif(module_import_skip, reason = 'cannot import IIRrational_test_data')
def test_std_4roots():
    X_scale = 100
    roots = [-0.8 * X_scale, 0 + 0.5j * X_scale, 0 - 0.5j * X_scale, -0.5 * X_scale]
    # print(standard.roots(standard.fromroots_lnG(roots, X_scale = X_scale)[0], X_scale = X_scale))

    c, lnG = standard.fromroots_lnG(roots, X_scale=X_scale)
    print(c, np.exp(lnG))

    F = np.linspace(0, X_scale, X_scale)
    X = 1j * F

    y_main, y_main_lnG = standard.valfromroots_lnG(X, roots, X_scale=X_scale, lnG=lnG)
    y_poly, y_poly_lnG = standard.val_lnG(X, c, X_scale=X_scale, lnG=lnG)
    V, VlnG = standard.vander_lnG(X, len(c) - 1, X_scale=X_scale, lnG=lnG)
    y_vander = np.dot(V, c) * np.exp(VlnG)

    # from wield.iirrational.utilities.mpl import mplfigB
    # axB = mplfigB()
    # axB.ax0.plot(F, abs(y_main))
    # axB.ax0.plot(F, abs(y_poly))
    # axB.ax0.plot(F, abs(y_vander))

    np.testing.assert_allclose(y_vander, y_poly * np.exp(y_poly_lnG))
    np.testing.assert_allclose(y_main * np.exp(y_main_lnG) / y_vander, 1)
    np.testing.assert_allclose(y_main * np.exp(y_main_lnG), y_poly * np.exp(y_poly_lnG))
    return


def test_std_3roots():
    X_scale = 100
    roots = [-0.8 * X_scale, 0 + 0.5j * X_scale, 0 - 0.5j * X_scale]
    print(
        standard.roots(
            standard.fromroots_lnG(roots, X_scale=X_scale)[0], X_scale=X_scale
        )
    )

    c, lnG = standard.fromroots_lnG(roots, X_scale=X_scale)

    F = np.linspace(0, X_scale, X_scale)
    X = 1j * F

    y_main, y_main_lnG = standard.valfromroots_lnG(X, roots, X_scale=X_scale, lnG=lnG)
    y_poly, y_poly_lnG = standard.val_lnG(X, c, X_scale=X_scale, lnG=lnG)
    V, VlnG = standard.vander_lnG(X, len(c) - 1, X_scale=X_scale, lnG=lnG)
    y_vander = np.dot(V, c) * np.exp(VlnG)

    # from wield.iirrational.utilities.mpl import mplfigB
    # axB = mplfigB()
    # axB.ax0.plot(F, abs(y_main))
    # axB.ax0.plot(F, abs(y_poly))
    # axB.ax0.plot(F, abs(y_vander))

    np.testing.assert_allclose(y_main * np.exp(y_main_lnG), y_poly * np.exp(y_poly_lnG))
    np.testing.assert_allclose(y_main * np.exp(y_main_lnG), y_vander)
    return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""
from wield.bunch import Bunch

from ..codings_cmn import (
    CodingTypeZ,
    Ipi,
    I2pi,
    BranchCutAmbiguity,
    CodingGain,
    CodingDelay,
)

from ..MRF import MultiReprFilterZ
from .cplx_real_imag import CodingRI

from .cplx_frequency_BW import CodingFBW
from .cplx_freq_nl_BW import CodingFnlBW
from .cplx_nl_freq_nl_BW import CodingnlFnlBW

from .cplx_frequency_amplitude import CodingFA, CodingFAUX
from .cplx_frequency_nl_amp import CodingFnlA
from .cplx_nl_freq_nl_amp import CodingnlFnlA

from .real_amplitude import CodingRealAmp, CodingRealAmpUX
from .real_BW import CodingRealBW

from .real_nl_amplitude import CodingRealnlA
from .real_nl_BW import CodingRealnlBW

from .cplx_sos import CodingSOS, CodingSOSMirror
from .multi_order_section import CodingMOS

import sys

# get the current module object to provide a reference
_module = sys.modules[__name__]


def is_unstable_Z(root):
    return abs(root) >= 1


_common = dict(
    module=_module,
    representation="Z",
    mrf_default=MultiReprFilterZ,
    is_unstable=is_unstable_Z,
)


coding_maps = Bunch(
    nlFBW=Bunch(
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingRealnlBW,
        num_c=CodingnlFnlBW,
        den_r=CodingRealnlBW,
        den_c=CodingnlFnlBW,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    nlFBW_safe=Bunch(
        # gain    = CodingGainNLDelay,
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingRealnlBW,
        num_c=CodingnlFnlBW,
        den_r=CodingRealnlBW,
        den_c=CodingnlFnlBW,
        num_r_u=CodingRealBW,
        num_c_u=CodingFBW,
        den_r_u=CodingRealBW,
        den_c_u=CodingFBW,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    nlFA=Bunch(
        # gain    = CodingGainNLDelay,
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingRealnlA,
        num_c=CodingnlFnlA,
        den_r=CodingRealnlA,
        den_c=CodingnlFnlA,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    nlFA_safe=Bunch(
        # gain    = CodingGainNLDelay,
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingRealnlA,
        num_c=CodingnlFnlA,
        den_r=CodingRealnlA,
        den_c=CodingnlFnlA,
        num_r_u=CodingRealAmp,
        num_c_u=CodingFA,
        den_r_u=CodingRealAmp,
        den_c_u=CodingFA,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    FA=Bunch(
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingRealAmp,
        num_c=CodingFA,
        den_r=CodingRealAmp,
        den_c=CodingFA,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    FBW=Bunch(
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingRealBW,
        num_c=CodingFBW,
        den_r=CodingRealBW,
        den_c=CodingFBW,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    RI=Bunch(
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingRealAmp,
        num_c=CodingRI,
        den_r=CodingRealAmp,
        den_c=CodingRI,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=1,
        den_collect_r=1,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
    SOS=Bunch(
        gain=CodingGain,
        delay=CodingDelay,
        num_r=CodingSOS,
        num_c=CodingSOS,
        den_r=CodingSOS,
        den_c=CodingSOS,
        num_r_u=None,
        num_c_u=None,
        den_r_u=None,
        den_c_u=None,
        num_collect_r=2,
        den_collect_r=2,
        num_collect_c=1,
        den_collect_c=1,
        **_common
    ),
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: © 2021 Massachusetts Institute of Technology.
# SPDX-FileCopyrightText: © 2021 Lee McCuller <mcculler@caltech.edu>
# NOTICE: authors should document their contributions in concisely in NOTICE
# with details inline in source files, comments, and docstrings.
"""
"""

import sympy

print("Sympy version: ", sympy.__version__)

# this makes the notebook sexy
sympy.init_printing(use_latex="mathjax")

from IPython.display import (
    display,
    display_pretty,
    display_html,
    display_jpeg,
    display_png,
    display_json,
    display_latex,
    display_svg,
    clear_output,
    Latex,
)
